"""migrate records from old format to new.

Revision ID: 71b84ccc31bb
Revises: f2772c2c64a6
Create Date: 2018-02-14 12:04:34.688790

"""

# revision identifiers, used by Alembic.
revision = '71b84ccc31bb'
down_revision = 'f2772c2c64a6'

from alembic import op
# These are the "lightweight" SQL expression versions (not using metadata):
from sqlalchemy.sql.expression import table, column, select, update
from sqlalchemy.sql.sqltypes import Integer, Text

import json
import requests

from waiverdb.api_v1 import get_resultsdb_result


def convert_id_to_subject_and_testcase(result_id):
    try:
        result = get_resultsdb_result(result_id)
    except requests.HTTPError as e:
        if e.response.status_code == 404:
            raise RuntimeError('Result id %s not found in Resultsdb' % (result_id)) from e
        else:
            raise RuntimeError('Failed looking up result in Resultsdb') from e
    except Exception as e:
        raise RuntimeError('Failed looking up result in Resultsdb') from e
    if 'original_spec_nvr' in result['data']:
        subject = {'original_spec_nvr': result['data']['original_spec_nvr'][0]}
    else:
        if result['data']['type'][0] == 'koji_build' or \
           result['data']['type'][0] == 'bodhi_update':
            SUBJECT_KEYS = ['item', 'type']
            subject = dict([(k, v[0]) for k, v in result['data'].items()
                            if k in SUBJECT_KEYS])
        else:
            raise RuntimeError('Unable to determine subject for result id %s' % (result_id))
    testcase = result['testcase']['name']
    return (subject, testcase)


def upgrade():
    # Lightweight table definition for producing UPDATE queries.
    waiver_table = table('waiver',
                         column('id', type_=Integer),
                         column('result_id', type_=Integer),
                         column('subject', type_=Text),
                         column('testcase', type_=Text))
    # Get a session asociated with the alembic upgrade operation.
    connection = op.get_bind()
    rows = connection.execute(select([waiver_table.c.id, waiver_table.c.result_id]))
    for waiver_id, result_id in rows:
        subject, testcase = convert_id_to_subject_and_testcase(result_id)
        connection.execute(update(waiver_table)
                           .where(waiver_table.c.id == waiver_id)
                           .values(subject=json.dumps(subject), testcase=testcase))


def downgrade():
    # It shouldn't be possible to downgrade this change.
    # Because the result_id field will not be populated with data anymore.
    # If the user tries to downgrade "result_id" should be not null once again
    # like in the old version of the schema, but the value is not longer available
    raise RuntimeError('Irreversible migration')



# SPDX-License-Identifier: GPL-2.0+

import os

try:
    from urllib.parse import urlparse, urlunsplit
except ImportError:
    from urlparse import urlparse, urlunsplit

from flask import Flask, current_app
from flask_cors import CORS
from flask_migrate import Migrate
from sqlalchemy import event
from sqlalchemy.exc import ProgrammingError
import requests

from waiverdb.events import publish_new_waiver
from waiverdb.logger import init_logging
from waiverdb.api_v1 import api_v1
from waiverdb.models import db
from waiverdb.utils import json_error
from flask_oidc import OpenIDConnect
from werkzeug.exceptions import default_exceptions
from waiverdb.monitor import db_hook_event_listeners


def enable_cors(app):
    """
    Enables CORS headers.
    """
    # backward compatibility with old CORS_URL option
    cors_url = app.config.get('CORS_URL')
    if cors_url:
        app.config['CORS_ORIGINS'] = cors_url

    CORS(app)


def load_config(app):
    # Load default config, then override that with a config file
    if os.getenv('DEV') == 'true':
        default_config_obj = 'waiverdb.config.DevelopmentConfig'
        default_config_file = os.getcwd() + '/conf/settings.py'
        silent = True
    elif os.getenv('TEST') == 'true':
        default_config_obj = 'waiverdb.config.TestingConfig'
        default_config_file = os.getcwd() + '/conf/settings.py'
        silent = True
    elif os.getenv('DOCS') == 'true':
        default_config_obj = 'waiverdb.config.DevelopmentConfig'
        default_config_file = os.getcwd() + '/conf/settings.py'
        silent = True
    else:
        default_config_obj = 'waiverdb.config.ProductionConfig'
        default_config_file = '/etc/waiverdb/settings.py'
        silent = False

    app.config.from_object(default_config_obj)
    config_file = os.environ.get('WAIVERDB_CONFIG', default_config_file)
    app.config.from_pyfile(config_file, silent=silent)

    # Allow overriding only DATABASE_URI for tests.
    if os.getenv('TEST') == 'true':
        db_uri = app.config['DATABASE_URI']
        app.config.from_object(default_config_obj)
        app.config['DATABASE_URI'] = db_uri

    if os.environ.get('SECRET_KEY'):
        app.config['SECRET_KEY'] = os.environ['SECRET_KEY']


def populate_db_config(app):
    # Take the application-level DATABASE_URI setting, plus (optionally)
    # a DATABASE_PASSWORD from the environment, and munge them together into
    # the SQLALCHEMY_DATABASE_URI setting which is obeyed by Flask-SQLAlchemy.
    dburi = app.config['DATABASE_URI']
    if os.environ.get('DATABASE_PASSWORD'):
        parsed = urlparse(dburi)
        netloc = '{}:{}@{}'.format(parsed.username,
                                   os.environ['DATABASE_PASSWORD'],
                                   parsed.hostname)
        if parsed.port:
            netloc += ':{}'.format(parsed.port)
        dburi = urlunsplit(
            (parsed.scheme, netloc, parsed.path, parsed.query, parsed.fragment))
    if app.config['SHOW_DB_URI']:
        app.logger.debug('using DBURI: %s', dburi)
    app.config['SQLALCHEMY_DATABASE_URI'] = dburi


# applicaiton factory http://flask.pocoo.org/docs/0.12/patterns/appfactories/
def create_app(config_obj=None):
    app = Flask(__name__)
    if config_obj:
        app.config.from_object(config_obj)
    else:
        load_config(app)
    if app.config['PRODUCTION'] and app.secret_key == 'replace-me-with-something-random':
        raise Warning("You need to change the app.secret_key value for production")

    # register error handlers
    for code in default_exceptions.keys():
        app.register_error_handler(code, json_error)
    app.register_error_handler(requests.ConnectionError, json_error)
    app.register_error_handler(requests.Timeout, json_error)

    populate_db_config(app)
    if app.config['AUTH_METHOD'] == 'OIDC':
        app.oidc = OpenIDConnect(app)
    # initialize logging
    init_logging(app)
    # initialize db
    db.init_app(app)
    # initialize db migrations
    migrations_dir = os.path.join(os.path.abspath(os.path.dirname(__file__)),
                                  'migrations')
    Migrate(app, db, directory=migrations_dir)
    # register blueprints
    app.register_blueprint(api_v1, url_prefix="/api/v1.0")
    app.add_url_rule('/healthcheck', view_func=healthcheck)
    register_event_handlers(app)

    # initialize DB event listeners from the monitor module
    app.before_first_request(db_hook_event_listeners)

    enable_cors(app)

    return app


def healthcheck():
    """
    Request handler for performing an application-level health check. This is
    not part of the published API, it is intended for use by OpenShift or other
    monitoring tools.

    Returns a 200 response if the application is alive and able to serve requests.
    """
    try:
        db.session.execute("SELECT 1 FROM waiver LIMIT 0").fetchall()
    except ProgrammingError:
        current_app.logger.exception('Healthcheck failed on DB query.')
        raise RuntimeError('Unable to communicate with database.')

    return ('Health check OK', 200, [('Content-Type', 'text/plain')])


def register_event_handlers(app):
    """
    Register SQLAlchemy event handlers with the application's session factory.

    Args:
        app (flask.Flask): The Flask object with the configured scoped session
            attached as the ``session`` attribute.
    """
    if app.config['MESSAGE_BUS_PUBLISH']:
        # A workaround for https://github.com/mitsuhiko/flask-sqlalchemy/pull/364
        # can be removed after python-flask-sqlalchemy is upgraded to 2.2
        from flask_sqlalchemy import SignallingSession
        event.listen(SignallingSession, 'after_commit', publish_new_waiver)



# SPDX-License-Identifier: GPL-2.0+

from waiverdb.models import Waiver


def create_waiver(session, subject_type, subject_identifier, testcase, username, product_version,
                  waived=True, comment=None, proxied_by=None, scenario=None):
    waiver = Waiver(subject_type, subject_identifier, testcase, username, product_version,
                    waived, comment, proxied_by, scenario)
    session.add(waiver)
    session.flush()
    return waiver



"""(OBSOLETED) Change waiver.subject column type to JSONB

Revision ID: 1797bff52162
Revises: ed43eb9b221c
Create Date: 2018-02-23 16:37:44.190560

"""

# revision identifiers, used by Alembic.
revision = '1797bff52162'
down_revision = 'ed43eb9b221c'


def upgrade():
    # This migration originally attempted to use the JSONB column type,
    # which would fail on Postgres 9.2.
    # So this is now replaced by ce8a1351ecdc_change_subject_to_json.py
    # however we keep this as a no-op, to avoid breaking any Waiverdb
    # deployment where this migration had successfully been applied.
    pass


def downgrade():
    pass



# SPDX-License-Identifier: GPL-2.0+
"""
This module contains a set of `SQLAlchemy event`_ hooks.

To use these hooks, you must register them with SQLAlchemy
using the :func:`sqlalchemy.event.listen` function.

.. _SQLALchemy events:
    https://docs.sqlalchemy.org/en/latest/orm/events.html
"""

import logging
import time

from flask_restful import marshal
import stomp
import json
import waiverdb.monitor as monitor

from fedora_messaging.api import Message, publish
from fedora_messaging.exceptions import PublishReturned, ConnectionException
from flask import current_app
from waiverdb.fields import waiver_fields
from waiverdb.models import Waiver
from waiverdb.utils import stomp_connection

_log = logging.getLogger(__name__)

MAX_STOMP_RETRY = 3
STOMP_RETRY_DELAY_SECONDS = 5


def _send_stomp_message(session):
    with stomp_connection() as conn:
        stomp_configs = current_app.config.get('STOMP_CONFIGS')
        for row in session.identity_map.values():
            monitor.messaging_tx_to_send_counter.inc()
            if not isinstance(row, Waiver):
                continue
            _log.debug('Publishing a message for %r', row)
            msg = json.dumps(marshal(row, waiver_fields))
            kwargs = dict(body=msg, headers={}, destination=stomp_configs['destination'])
            if stomp.__version__[0] < 4:
                kwargs['message'] = kwargs.pop('body')  # On EL7, different sig.
            try:
                conn.send(**kwargs)
                monitor.messaging_tx_sent_ok_counter.inc()
            except Exception:
                _log.exception('Couldn\'t publish message via stomp')
                monitor.messaging_tx_failed_counter.inc()
                raise


def _send_stomp_message_with_retry(session, max_retry, retry_delay):
    for i in range(max_retry):
        time.sleep(i * retry_delay)
        try:
            _send_stomp_message(session)
        except stomp.exception.StompException:
            _log.exception('Failed to send message (try %s/%s)', i + 1, max_retry)
        else:
            break


def publish_new_waiver(session):
    """
    A post-commit event hook that emits messages to a message bus. The messages
    can be published by either fedora-messaging or stomp.

    This event is designed to be registered with a session factory::

        >>> from sqlalchemy.event import listen
        >>> listen(MyScopedSession, 'after_commit', publish_new_waiver)

    The emitted message will look like::

        {
          "username": "jcline",
          "i": 4,
          "timestamp": 1489686124,
          "msg_id": "2017-80e46243-e6f5-46df-8dcd-4d17809eb298",
          "topic": "org.fedoraproject.dev.waiverdb.waiver.new",
          "msg": {
            "comment": "Because I said so",
            "username": "http://jcline.id.fedoraproject.org/",
            "waived": true,
            "timestamp": "2017-03-16T17:42:04.209638",
            "product_version": "Satellite 6.3",
            "subject": "{\"a.nice.example\": \"this-is-a-really-nice-example\"}",
            "testcase": "t.e.s.t.case",
            "proxied_by": null,
            "id": 15
          }
        }

    Args:
        session (sqlalchemy.orm.Session): The session that was committed to the
            database. This session is not active and cannot emit SQL.

    """
    _log.debug('The publish_new_waiver SQLAlchemy event has been activated (%r)',
               current_app.config['MESSAGE_PUBLISHER'])

    if current_app.config['MESSAGE_PUBLISHER'] == 'stomp':
        max_retry = current_app.config.get('MAX_STOMP_RETRY', MAX_STOMP_RETRY)
        retry_delay = current_app.config.get('STOMP_RETRY_DELAY_SECONDS', STOMP_RETRY_DELAY_SECONDS)
        _send_stomp_message_with_retry(session, max_retry=max_retry, retry_delay=retry_delay)

    elif current_app.config['MESSAGE_PUBLISHER'] == 'fedmsg':
        for row in session.identity_map.values():
            monitor.messaging_tx_to_send_counter.inc()
            if not isinstance(row, Waiver):
                continue
            _log.debug('Publishing a message for %r', row)
            try:
                msg = Message(
                    topic='waiverdb.waiver.new',
                    body=marshal(row, waiver_fields)
                )
                publish(msg)
                monitor.messaging_tx_sent_ok_counter.inc()
            except PublishReturned as e:
                _log.exception('Fedora Messaging broker rejected message %s: %s', msg.id, e)
                monitor.messaging_tx_failed_counter.inc()
            except ConnectionException as e:
                _log.exception('Error sending message %s: %s', msg.id, e)
                monitor.messaging_tx_failed_counter.inc()
                raise

    elif current_app.config['MESSAGE_PUBLISHER'] is None:
        _log.info('No message published.  MESSAGE_PUBLISHER disabled.')
        monitor.messaging_tx_stopped_counter.inc()

    else:
        _log.warning('Unhandled MESSAGE_PUBLISHER %r', current_app.config['MESSAGE_PUBLISHER'])
        monitor.messaging_tx_failed_counter.inc()



# SPDX-License-Identifier: GPL-2.0+

import logging
import re
from fnmatch import fnmatch

from werkzeug.exceptions import (
    BadGateway,
    InternalServerError,
    Unauthorized,
)

log = logging.getLogger(__name__)


def get_group_membership(ldap, user, con, ldap_search):
    try:
        results = con.search_s(
            ldap_search['BASE'], ldap.SCOPE_SUBTREE,
            ldap_search.get('SEARCH_STRING', '(memberUid={user})').format(user=user), ['cn']
        )
        return [group[1]['cn'][0].decode('utf-8') for group in results]
    except KeyError:
        log.exception('LDAP_SEARCHES parameter should contain the BASE key')
        raise InternalServerError('LDAP_SEARCHES parameter should contain the BASE key')
    except ldap.SERVER_DOWN:
        log.exception('The LDAP server is not reachable.')
        raise BadGateway('The LDAP server is not reachable.')
    except ldap.LDAPError:
        log.exception('Some error occurred initializing the LDAP connection.')
        raise Unauthorized('Some error occurred initializing the LDAP connection.')


def match_testcase_permissions(testcase, permissions):
    for permission in permissions:
        if "testcases" in permission:
            testcase_match = any(
                fnmatch(testcase, testcase_pattern)
                for testcase_pattern in permission["testcases"]
            )
        elif "_testcase_regex_pattern" in permission:
            testcase_match = re.search(
                permission["_testcase_regex_pattern"], testcase)
        else:
            continue

        if testcase_match:
            yield permission


def verify_authorization(user, testcase, permissions, ldap_host, ldap_searches):
    if not (ldap_host and ldap_searches):
        raise InternalServerError(('LDAP_HOST and LDAP_SEARCHES also need to be defined '
                                   'if PERMISSIONS is defined.'))

    allowed_groups = []
    for permission in match_testcase_permissions(testcase, permissions):
        if user in permission.get('users', []):
            return True
        allowed_groups += permission.get('groups', [])

    try:
        import ldap
    except ImportError:
        raise InternalServerError(('If PERMISSIONS is defined, '
                                   'python-ldap needs to be installed.'))

    try:
        con = ldap.initialize(ldap_host)
    except ldap.LDAPError:
        log.exception('Some error occurred initializing the LDAP connection.')
        raise Unauthorized('Some error occurred initializing the LDAP connection.')
    group_membership = set()

    for cur_ldap_search in ldap_searches:
        group_membership.update(
            get_group_membership(ldap, user, con, cur_ldap_search)
        )
        if group_membership & set(allowed_groups):
            return True

    if not group_membership:
        raise Unauthorized(f'Couldn\'t find user {user} in LDAP')

    raise Unauthorized(('You are not authorized to submit a waiver '
                        f'for the test case {testcase}'))



# SPDX-License-Identifier: GPL-2.0+

import datetime
import json

import pytest
from requests import ConnectionError, HTTPError
from mock import patch, Mock
from stomp.exception import StompException

from .utils import create_waiver
from waiverdb import __version__
from waiverdb.models import Waiver


@pytest.fixture
def mocked_get_user(username):
    with patch('waiverdb.auth.get_user', return_value=(username, {})):
        yield username


@pytest.fixture
def mocked_user():
    with patch('waiverdb.auth.get_user', return_value=('foo', {})):
        yield 'foo'


@pytest.fixture
def mocked_bodhi_user():
    with patch('waiverdb.auth.get_user', return_value=('bodhi', {})):
        yield 'bodhi'


@pytest.fixture
def mocked_resultsdb():
    with patch('waiverdb.api_v1.get_resultsdb_result') as mocked_resultsdb:
        yield mocked_resultsdb


def test_create_waiver(mocked_user, client, session):
    data = {
        'subject_type': 'koji_build',
        'subject_identifier': 'glibc-2.26-27.fc27',
        'testcase': 'testcase1',
        'product_version': 'fool-1',
        'waived': True,
        'comment': 'it broke',
    }
    r = client.post('/api/v1.0/waivers/', data=json.dumps(data),
                    content_type='application/json')
    res_data = json.loads(r.get_data(as_text=True))
    assert r.status_code == 201
    assert res_data['username'] == 'foo'
    assert res_data['subject'] == {'type': 'koji_build', 'item': 'glibc-2.26-27.fc27'}
    assert res_data['subject_type'] == 'koji_build'
    assert res_data['subject_identifier'] == 'glibc-2.26-27.fc27'
    assert res_data['testcase'] == 'testcase1'
    assert res_data['product_version'] == 'fool-1'
    assert res_data['waived'] is True
    assert res_data['comment'] == 'it broke'
    assert res_data['scenario'] is None


def test_create_waiver_with_subject(mocked_user, client, session):
    # 'subject' key was the API in Waiverdb < 0.11
    data = {
        'subject': {'type': 'koji_build', 'item': 'glibc-2.26-27.fc27'},
        'subject_identifier': 'glibc-2.26-27.fc27',
        'testcase': 'dist.rpmdeplint',
        'product_version': 'fedora-27',
        'waived': True,
        'comment': 'it really broke',
    }
    r = client.post('/api/v1.0/waivers/', data=json.dumps(data),
                    content_type='application/json')
    res_data = json.loads(r.get_data(as_text=True))
    assert r.status_code == 201
    assert res_data['username'] == 'foo'
    assert res_data['subject'] == {'type': 'koji_build', 'item': 'glibc-2.26-27.fc27'}
    assert res_data['subject_type'] == 'koji_build'
    assert res_data['subject_identifier'] == 'glibc-2.26-27.fc27'
    assert res_data['testcase'] == 'dist.rpmdeplint'
    assert res_data['product_version'] == 'fedora-27'
    assert res_data['waived'] is True
    assert res_data['comment'] == 'it really broke'


def test_create_waiver_with_result_id(mocked_user, mocked_resultsdb, client, session):
    mocked_resultsdb.return_value = {
        'data': {
            'type': ['koji_build'],
            'item': ['somebuild'],
            'scenario': ['somescenario'],
        },
        'testcase': {'name': 'sometest'}
    }

    # 'result_id' key was the API in Waiverdb < 0.6
    data = {
        'result_id': 123,
        'product_version': 'fool-1',
        'waived': True,
        'comment': 'it broke',
    }
    r = client.post('/api/v1.0/waivers/', data=json.dumps(data),
                    content_type='application/json')
    res_data = json.loads(r.get_data(as_text=True))
    assert r.status_code == 201
    assert res_data['username'] == 'foo'
    assert res_data['scenario'] == 'somescenario'
    assert res_data['subject'] == {'type': 'koji_build', 'item': 'somebuild'}
    assert res_data['subject_type'] == 'koji_build'
    assert res_data['subject_identifier'] == 'somebuild'
    assert res_data['testcase'] == 'sometest'
    assert res_data['product_version'] == 'fool-1'
    assert res_data['waived'] is True
    assert res_data['comment'] == 'it broke'


def test_create_waiver_with_result_for_original_spec_nvr(
        mocked_user, mocked_resultsdb, client, session):
    mocked_resultsdb.return_value = {
        'data': {
            'original_spec_nvr': ['somedata'],
        },
        'testcase': {'name': 'sometest'}
    }

    data = {
        'result_id': 123,
        'product_version': 'fool-1',
        'waived': True,
        'comment': 'it broke',
    }
    r = client.post('/api/v1.0/waivers/', data=json.dumps(data),
                    content_type='application/json')
    res_data = json.loads(r.get_data(as_text=True))
    assert r.status_code == 201
    assert res_data['username'] == 'foo'
    assert res_data['subject'] == {'type': 'koji_build', 'item': 'somedata'}
    assert res_data['subject_type'] == 'koji_build'
    assert res_data['subject_identifier'] == 'somedata'
    assert res_data['testcase'] == 'sometest'
    assert res_data['product_version'] == 'fool-1'
    assert res_data['waived'] is True
    assert res_data['comment'] == 'it broke'


def test_create_waiver_without_comment(mocked_user, client, session):
    data = {
        'subject_type': 'koji_build',
        'subject_identifier': 'glibc-2.26-27.fc27',
        'product_version': 'fool-1',
        'waived': True,
    }
    r = client.post('/api/v1.0/waivers/', data=json.dumps(data),
                    content_type='application/json')
    res_data = json.loads(r.get_data(as_text=True))
    assert r.status_code == 400
    res_data = json.loads(r.get_data(as_text=True))
    assert res_data['message']['comment'] == 'Missing required parameter in the JSON body'


def test_create_waiver_with_scenario(mocked_user, client, session):
    data = {
        'subject_type': 'koji_build',
        'subject_identifier': 'glibc-2.26-27.fc27',
        'testcase': 'testcase1',
        'product_version': 'fool-1',
        'scenario': 'scenario1',
        'waived': True,
        'comment': 'it broke',
    }
    r = client.post('/api/v1.0/waivers/', data=json.dumps(data),
                    content_type='application/json')
    res_data = json.loads(r.get_data(as_text=True))
    assert r.status_code == 201
    assert res_data['username'] == 'foo'
    assert res_data['subject'] == {'type': 'koji_build', 'item': 'glibc-2.26-27.fc27'}
    assert res_data['subject_type'] == 'koji_build'
    assert res_data['subject_identifier'] == 'glibc-2.26-27.fc27'
    assert res_data['testcase'] == 'testcase1'
    assert res_data['product_version'] == 'fool-1'
    assert res_data['waived'] is True
    assert res_data['comment'] == 'it broke'
    assert res_data['scenario'] == 'scenario1'


def test_create_waiver_with_unknown_result_id(mocked_user, mocked_resultsdb, client, session):
    mocked_resultsdb.side_effect = HTTPError(response=Mock(status=404))
    data = {
        'result_id': 123,
        'product_version': 'fool-1',
        'waived': True,
        'comment': 'it broke',
    }
    mocked_resultsdb.return_value.status_code = 404
    r = client.post('/api/v1.0/waivers/', data=json.dumps(data),
                    content_type='application/json')
    res_data = json.loads(r.get_data(as_text=True))
    assert res_data['message'].startswith('Failed looking up result in Resultsdb:')


def test_create_waiver_with_no_testcase(mocked_user, client):
    data = {
        'subject_type': 'koji_build',
        'subject_identifier': 'glibc-2.26-27.fc27',
        'waived': True,
        'product_version': 'the-best',
        'comment': 'it broke',
    }
    r = client.post('/api/v1.0/waivers/', data=json.dumps(data),
                    content_type='application/json')
    res_data = json.loads(r.get_data(as_text=True))
    assert r.status_code == 400
    assert 'Missing required parameter in the JSON body' in res_data['message']['testcase']


def test_create_waiver_with_malformed_subject(mocked_user, client):
    data = {
        'subject': 'asd',
        'testcase': 'qqq',
    }
    r = client.post('/api/v1.0/waivers/', data=json.dumps(data),
                    content_type='application/json')
    res_data = json.loads(r.get_data(as_text=True))
    assert r.status_code == 400
    assert 'Must be a valid dict' in res_data['message']['subject']


def test_non_superuser_cannot_create_waiver_for_other_users(mocked_user, client):
    data = {
        'subject_type': 'koji_build',
        'subject_identifier': 'glibc-2.26-27.fc27',
        'testcase': 'testcase1',
        'product_version': 'fool-1',
        'waived': True,
        'comment': 'it broke',
        'username': 'bar',
    }
    r = client.post('/api/v1.0/waivers/', data=json.dumps(data),
                    content_type='application/json')
    res_data = json.loads(r.get_data(as_text=True))
    assert r.status_code == 403
    assert 'user foo does not have the proxyuser ability' == res_data['message']


def test_superuser_can_create_waiver_for_other_users(mocked_bodhi_user, client, session):
    data = {
        'subject_type': 'koji_build',
        'subject_identifier': 'glibc-2.26-27.fc27',
        'testcase': 'testcase1',
        'product_version': 'fool-1',
        'waived': True,
        'comment': 'it broke',
        'username': 'bar',
    }
    r = client.post('/api/v1.0/waivers/', data=json.dumps(data),
                    content_type='application/json')
    res_data = json.loads(r.get_data(as_text=True))
    assert r.status_code == 201
    # a waiver should be created for bar by bodhi
    assert res_data['username'] == 'bar'
    assert res_data['proxied_by'] == mocked_bodhi_user


def test_get_waiver(client, session):
    # create a new waiver
    waiver = create_waiver(session, subject_type='koji_build',
                           subject_identifier='glibc-2.26-27.fc27',
                           testcase='testcase1', username='foo',
                           product_version='foo-1', comment='bla bla bla')
    r = client.get('/api/v1.0/waivers/%s' % waiver.id)
    res_data = json.loads(r.get_data(as_text=True))
    assert r.status_code == 200
    assert res_data['username'] == waiver.username
    assert res_data['subject'] == {'type': 'koji_build', 'item': 'glibc-2.26-27.fc27'}
    assert res_data['subject_type'] == waiver.subject_type
    assert res_data['subject_identifier'] == waiver.subject_identifier
    assert res_data['testcase'] == waiver.testcase
    assert res_data['product_version'] == waiver.product_version
    assert res_data['waived'] is True
    assert res_data['comment'] == waiver.comment
    assert res_data['scenario'] is None


def test_get_waiver_by_scenario(client, session):
    scenario_name = 'scenario19'
    # create a new waiver
    waiver1 = create_waiver(
        session, subject_type='koji_build', subject_identifier='glibc-2.26-27.fc27',
        testcase='testcase1', username='foo', product_version='foo-1', comment='bla bla bla',
        scenario=scenario_name
    )
    create_waiver(
        session, subject_type='bodhi_update', subject_identifier='glibc-2.26-27.fc27',
        testcase='testcase1', username='foo', product_version='foo-1', comment='bla bla bla',
        scenario='yyyyyy'
    )
    r = client.get(f'/api/v1.0/waivers/?scenario={scenario_name}')
    assert r.status_code == 200
    res_data = json.loads(r.get_data(as_text=True))
    assert len(res_data['data']) == 1
    assert res_data['data'][0]['username'] == waiver1.username
    assert res_data['data'][0]['subject'] == {'type': 'koji_build', 'item': 'glibc-2.26-27.fc27'}
    assert res_data['data'][0]['subject_type'] == waiver1.subject_type
    assert res_data['data'][0]['subject_identifier'] == waiver1.subject_identifier
    assert res_data['data'][0]['testcase'] == waiver1.testcase
    assert res_data['data'][0]['product_version'] == waiver1.product_version
    assert res_data['data'][0]['waived'] is True
    assert res_data['data'][0]['comment'] == waiver1.comment
    assert res_data['data'][0]['scenario'] == scenario_name
    assert res_data['prev'] is None
    assert res_data['next'] is None
    assert res_data['first'] == res_data['last']
    assert res_data['last'] == f'http://localhost/api/v1.0/waivers/?page=1&scenario={scenario_name}'


def test_404_for_nonexistent_waiver(client, session):
    r = client.get('/api/v1.0/waivers/foo')
    assert r.status_code == 404
    res_data = json.loads(r.get_data(as_text=True))
    assert 'The requested URL was not found on the server' in res_data['message']


@patch('waiverdb.api_v1.AboutResource.get', side_effect=ConnectionError)
def test_500(mocked, client, session):
    r = client.get('/api/v1.0/about')
    assert r.status_code == 500
    res_data = json.loads(r.get_data(as_text=True))
    assert res_data['message'] == ''


def test_get_waivers(client, session):
    for i in range(0, 10):
        create_waiver(session, subject_type='koji_build', subject_identifier="%d" % i,
                      testcase="case %d" % i, username='foo %d' % i,
                      product_version='foo-%d' % i, comment='bla bla bla')
    r = client.get('/api/v1.0/waivers/')
    res_data = json.loads(r.get_data(as_text=True))
    assert r.status_code == 200
    assert len(res_data['data']) == 10


def test_pagination_waivers(client, session):
    for i in range(0, 30):
        create_waiver(session, subject_type='koji_build', subject_identifier="%d" % i,
                      testcase="case %d" % i, username='foo %d' % i,
                      product_version='foo-%d' % i, comment='bla bla bla')
    r = client.get('/api/v1.0/waivers/?page=2')
    res_data = json.loads(r.get_data(as_text=True))
    assert r.status_code == 200
    assert len(res_data['data']) == 10
    assert '/waivers/?page=1' in res_data['prev']
    assert '/waivers/?page=3' in res_data['next']
    assert '/waivers/?page=1' in res_data['first']
    assert '/waivers/?page=3' in res_data['last']


def test_obsolete_waivers_are_excluded_by_default(client, session):
    create_waiver(session, subject_type='koji_build',
                  subject_identifier='glibc-2.26-27.fc27',
                  testcase='testcase1', username='foo',
                  product_version='foo-1')
    new_waiver = create_waiver(session, subject_type='koji_build',
                               subject_identifier='glibc-2.26-27.fc27',
                               testcase='testcase1', username='foo',
                               product_version='foo-1', waived=False)
    r = client.get('/api/v1.0/waivers/')
    res_data = json.loads(r.get_data(as_text=True))
    assert r.status_code == 200
    assert len(res_data['data']) == 1
    assert res_data['data'][0]['id'] == new_waiver.id
    assert res_data['data'][0]['waived'] == new_waiver.waived


def test_get_obsolete_waivers(client, session):
    old_waiver = create_waiver(session, subject_type='koji_build',
                               subject_identifier='glibc-2.26-27.fc27',
                               testcase='testcase1', username='foo',
                               product_version='foo-1')
    new_waiver = create_waiver(session, subject_type='koji_build',
                               subject_identifier='glibc-2.26-27.fc27',
                               testcase='testcase1', username='foo',
                               product_version='foo-1', waived=False)
    r = client.get('/api/v1.0/waivers/?include_obsolete=1')
    res_data = json.loads(r.get_data(as_text=True))
    assert r.status_code == 200
    assert len(res_data['data']) == 2
    assert res_data['data'][0]['id'] == new_waiver.id
    assert res_data['data'][1]['id'] == old_waiver.id


def test_obsolete_waivers_with_different_product_version(client, session):
    old_waiver = create_waiver(session, subject_type='koji_build',
                               subject_identifier='glibc-2.26-27.fc27',
                               testcase='testcase1', username='foo',
                               product_version='foo-1')
    new_waiver = create_waiver(session, subject_type='koji_build',
                               subject_identifier='glibc-2.26-27.fc27',
                               testcase='testcase1', username='foo',
                               product_version='foo-2')
    r = client.get('/api/v1.0/waivers/?include_obsolete=0')
    res_data = json.loads(r.get_data(as_text=True))
    assert r.status_code == 200
    assert len(res_data['data']) == 2
    assert res_data['data'][0]['id'] == new_waiver.id
    assert res_data['data'][1]['id'] == old_waiver.id


def test_obsolete_waivers_with_different_username(client, session):
    old_waiver = create_waiver(session, subject_type='koji_build',
                               subject_identifier='glibc-2.26-27.fc27',
                               testcase='testcase1', username='foo',
                               product_version='foo-1')
    new_waiver = create_waiver(session, subject_type='koji_build',
                               subject_identifier='glibc-2.26-27.fc27',
                               testcase='testcase1', username='bar',
                               product_version='foo-1')
    r = client.get('/api/v1.0/waivers/?include_obsolete=0')
    res_data = json.loads(r.get_data(as_text=True))
    assert r.status_code == 200
    assert len(res_data['data']) == 2
    assert res_data['data'][0]['id'] == new_waiver.id
    assert res_data['data'][1]['id'] == old_waiver.id


def test_filtering_waivers_by_subject_type(client, session):
    create_waiver(session, subject_type='koji_build', subject_identifier='glibc-2.26-27.fc27',
                  testcase='testcase', username='foo-1', product_version='foo-1')
    create_waiver(session, subject_type='bodhi_update', subject_identifier='FEDORA-2017-7e594f96bb',
                  testcase='testcase', username='foo-2', product_version='foo-1')

    r = client.get('/api/v1.0/waivers/?subject_type=bodhi_update')
    res_data = json.loads(r.get_data(as_text=True))
    assert r.status_code == 200
    assert len(res_data['data']) == 1
    assert res_data['data'][0]['subject_type'] == 'bodhi_update'


def test_filtering_waivers_by_subject_identifier(client, session):
    create_waiver(session, subject_type='koji_build', subject_identifier='glibc-2.26-27.fc27',
                  testcase='testcase', username='foo-1', product_version='foo-1')
    create_waiver(session, subject_type='koji_build', subject_identifier='kernel-4.15.17-300.fc27',
                  testcase='testcase', username='foo-2', product_version='foo-1')

    r = client.get('/api/v1.0/waivers/?subject_identifier=glibc-2.26-27.fc27')
    res_data = json.loads(r.get_data(as_text=True))
    assert r.status_code == 200
    assert len(res_data['data']) == 1
    assert res_data['data'][0]['subject_identifier'] == 'glibc-2.26-27.fc27'


def test_filtering_waivers_by_testcase(client, session):
    create_waiver(session, subject_type='koji_build', subject_identifier='glibc-2.26-27.fc27',
                  testcase='testcase1', username='foo-1', product_version='foo-1')
    create_waiver(session, subject_type='koji_build', subject_identifier='glibc-2.26-27.fc27',
                  testcase='testcase2', username='foo-2', product_version='foo-1')

    r = client.get('/api/v1.0/waivers/?testcase=testcase1')
    res_data = json.loads(r.get_data(as_text=True))
    assert r.status_code == 200
    assert len(res_data['data']) == 1
    assert res_data['data'][0]['testcase'] == 'testcase1'


def test_filtering_waivers_by_product_version(client, session):
    create_waiver(session, subject_type='koji_build', subject_identifier='glibc-2.26-27.fc27',
                  testcase='testcase1', username='foo-1', product_version='release-1')
    create_waiver(session, subject_type='koji_build', subject_identifier='kernel-4.15.17-300.fc27',
                  testcase='testcase2', username='foo-1', product_version='release-2')
    r = client.get('/api/v1.0/waivers/?product_version=release-1')
    res_data = json.loads(r.get_data(as_text=True))
    assert r.status_code == 200
    assert len(res_data['data']) == 1
    assert res_data['data'][0]['product_version'] == 'release-1'


def test_filtering_waivers_by_username(client, session):
    create_waiver(session, subject_type='koji_build', subject_identifier='glibc-2.26-27.fc27',
                  testcase='testcase1', username='foo', product_version='foo-1')
    create_waiver(session, subject_type='koji_build', subject_identifier='kernel-4.15.17-300.fc27',
                  testcase='testcase2', username='bar', product_version='foo-2')
    r = client.get('/api/v1.0/waivers/?username=foo')
    res_data = json.loads(r.get_data(as_text=True))
    assert r.status_code == 200
    assert len(res_data['data']) == 1
    assert res_data['data'][0]['username'] == 'foo'


def test_filtering_waivers_by_since(client, session):
    before1 = (datetime.datetime.utcnow() - datetime.timedelta(seconds=100)).isoformat()
    before2 = (datetime.datetime.utcnow() - datetime.timedelta(seconds=99)).isoformat()
    after = (datetime.datetime.utcnow() + datetime.timedelta(seconds=100)).isoformat()
    create_waiver(session, subject_type='koji_build', subject_identifier='glibc-2.26-27.fc27',
                  testcase='testcase1', username='foo', product_version='foo-1')
    r = client.get('/api/v1.0/waivers/?since=%s' % before1)
    res_data = json.loads(r.get_data(as_text=True))
    assert r.status_code == 200
    assert len(res_data['data']) == 1
    assert res_data['data'][0]['subject_identifier'] == 'glibc-2.26-27.fc27'
    assert res_data['data'][0]['testcase'] == 'testcase1'

    r = client.get('/api/v1.0/waivers/?since=%s,%s' % (before1, after))
    res_data = json.loads(r.get_data(as_text=True))
    assert r.status_code == 200
    assert len(res_data['data']) == 1
    assert res_data['data'][0]['subject_identifier'] == 'glibc-2.26-27.fc27'
    assert res_data['data'][0]['testcase'] == 'testcase1'

    r = client.get('/api/v1.0/waivers/?since=%s' % (after))
    res_data = json.loads(r.get_data(as_text=True))
    assert r.status_code == 200
    assert len(res_data['data']) == 0

    r = client.get('/api/v1.0/waivers/?since=%s,%s' % (before1, before2))
    res_data = json.loads(r.get_data(as_text=True))
    assert r.status_code == 200
    assert len(res_data['data']) == 0


def test_filtering_waivers_by_malformed_since(client, session):
    r = client.get('/api/v1.0/waivers/?since=123')
    res_data = json.loads(r.get_data(as_text=True))
    assert r.status_code == 400
    assert res_data['message']['since'] == \
        "time data '123' does not match format '%Y-%m-%dT%H:%M:%S.%f'"

    r = client.get('/api/v1.0/waivers/?since=%s,badend' % datetime.datetime.utcnow().isoformat())
    res_data = json.loads(r.get_data(as_text=True))
    assert r.status_code == 400
    assert res_data['message']['since'] == \
        "time data 'badend' does not match format '%Y-%m-%dT%H:%M:%S.%f'"

    r = client.get('/api/v1.0/waivers/?since=%s,too,many,commas'
                   % datetime.datetime.utcnow().isoformat())
    res_data = json.loads(r.get_data(as_text=True))
    assert r.status_code == 400
    assert res_data['message']['since'] == \
        "time data 'too,many,commas' does not match format '%Y-%m-%dT%H:%M:%S.%f'"


def test_filtering_waivers_by_proxied_by(client, session):
    create_waiver(session, subject_type='koji_build', subject_identifier='glibc-2.26-27.fc27',
                  testcase='testcase1', username='foo-1', product_version='foo-1',
                  proxied_by='bodhi')
    create_waiver(session, subject_type='koji_build', subject_identifier='kernel-4.15.17-300.fc27',
                  testcase='testcase2', username='foo-2', product_version='foo-1')
    r = client.get('/api/v1.0/waivers/?proxied_by=bodhi')
    res_data = json.loads(r.get_data(as_text=True))
    assert r.status_code == 200
    assert len(res_data['data']) == 1
    assert res_data['data'][0]['subject'] == {'type': 'koji_build', 'item': 'glibc-2.26-27.fc27'}
    assert res_data['data'][0]['testcase'] == 'testcase1'


def test_jsonp(client, session):
    waiver = create_waiver(session, subject_type='koji_build',
                           subject_identifier='glibc-2.26-27.fc27',
                           testcase='testcase1', username='foo', product_version='foo-1')
    r = client.get('/api/v1.0/waivers/%s?callback=jsonpcallback' % waiver.id)
    assert r.mimetype == 'application/javascript'
    assert 'jsonpcallback' in r.get_data(as_text=True)


def test_healthcheck(client):
    r = client.get('healthcheck')
    assert r.status_code == 200
    assert r.get_data(as_text=True) == 'Health check OK'


def test_filtering_waivers_with_post(client, session):
    filters = []
    for i in range(1, 51):
        filters.append({'subject_type': 'koji_build',
                        'subject_identifier': 'python2-2.7.14-%d.fc27' % i,
                        'testcase': 'case %d' % i})
        create_waiver(session, subject_type='koji_build',
                      subject_identifier='python2-2.7.14-%d.fc27' % i,
                      testcase='case %d' % i, username='person',
                      product_version='fedora-27', comment='bla bla bla')
    # Unrelated waiver which should not be included
    create_waiver(session, subject_type='koji_build',
                  subject_identifier='glibc-2.26-27.fc27',
                  testcase='dist.rpmdeplint', username='person',
                  product_version='fedora-27', comment='bla bla bla')
    r = client.post('/api/v1.0/waivers/+filtered',
                    data=json.dumps({'filters': filters}),
                    content_type='application/json')
    res_data = json.loads(r.get_data(as_text=True))
    assert r.status_code == 200
    assert len(res_data['data']) == 50
    assert all(w['subject_identifier'].startswith('python2-2.7.14') for w in res_data['data'])


def test_filtering_with_missing_filter(client, session):
    r = client.post('/api/v1.0/waivers/+filtered',
                    data=json.dumps({'somethingelse': 'what'}),
                    content_type='application/json')
    res_data = json.loads(r.get_data(as_text=True))
    assert r.status_code == 400
    assert res_data['message']['filters'] == 'Missing required parameter in the JSON body'


def test_waivers_by_subjects_and_testcases(client, session):
    """
    This tests that users can get waivers by sending a POST request with a long
    list of result subject/testcase.
    """
    results = []
    for i in range(1, 51):
        results.append({'subject': {'type': 'koji_build', 'item': '%d' % i},
                        'testcase': 'case %d' % i})
        create_waiver(session, subject_type='koji_build', subject_identifier="%d" % i,
                      testcase="case %d" % i, username='foo %d' % i,
                      product_version='foo-%d' % i, comment='bla bla bla')
    data = {
        'results': results
    }
    r = client.post('/api/v1.0/waivers/+by-subjects-and-testcases', data=json.dumps(data),
                    content_type='application/json')
    res_data = json.loads(r.get_data(as_text=True))
    assert r.status_code == 200
    assert len(res_data['data']) == 50
    subject_identifiers = []
    testcases = []
    for i in reversed(range(1, 51)):
        subject_identifiers.append('%d' % i)
        testcases.append('case %d' % i)
    assert [w['subject_identifier'] for w in res_data['data']] == subject_identifiers
    assert set([w['testcase'] for w in res_data['data']]) == set(testcases)
    assert all(w['username'].startswith('foo') for w in res_data['data'])
    assert all(w['product_version'].startswith('foo-') for w in res_data['data'])


@pytest.mark.parametrize("results", [
    [{'item': {'subject.test1': 'subject1'}}],  # Unexpected key
    [{'subject': 'subject1'}],  # Unexpected key type
])
def test_waivers_by_subjects_and_testcases_with_bad_results_parameter(client, session, results):
    data = {'results': results}
    r = client.post('/api/v1.0/waivers/+by-subjects-and-testcases', data=json.dumps(data),
                    content_type='application/json')
    res_data = json.loads(r.get_data(as_text=True))
    assert r.status_code == 400
    assert res_data['message']['results'] == \
        'Must be a list of dictionaries with "subject" and "testcase"'


def test_waivers_by_subjects_and_testcases_with_unrecognized_subject_type(client, session):
    create_waiver(session, subject_type='koji_build',
                  subject_identifier='python3-flask-0.12.2-1.fc29',
                  testcase='dist.rpmdeplint', username='person',
                  product_version='fedora-29', comment='bla bla bla')
    # This doesn't match any of the known subject types which we understand
    # for backwards compatibility. So if you tried to submit a waiver with a
    # subject like this, Waiverdb would reject it. But if the caller is just
    # *searching* and not *submitting* then instead of giving back an error
    # Waiverdb should just return an empty result set.
    data = {'results': [
        {'subject': {'item': 'python3-flask-0.12.2-1.fc29'}, 'testcase': 'dist.rpmdeplint'},
    ]}
    r = client.post('/api/v1.0/waivers/+by-subjects-and-testcases', data=json.dumps(data),
                    content_type='application/json')
    res_data = json.loads(r.get_data(as_text=True))
    assert r.status_code == 200
    assert res_data['data'] == []


@pytest.mark.parametrize("results", [
    [],
    [{}],
])
def test_waivers_by_subjects_and_testcases_with_empty_results_parameter(client, session, results):
    create_waiver(session, subject_type='koji_build', subject_identifier='glibc-2.26-27.fc27',
                  testcase='testcase1', username='foo-1', product_version='foo-1')
    data = {'results': results}
    r = client.post('/api/v1.0/waivers/+by-subjects-and-testcases', data=json.dumps(data),
                    content_type='application/json')
    res_data = json.loads(r.get_data(as_text=True))
    assert r.status_code == 200
    assert len(res_data['data']) == 1


def test_waivers_by_subjects_and_testcases_with_malformed_since(client, session):
    data = {'since': 123}
    r = client.post('/api/v1.0/waivers/+by-subjects-and-testcases', data=json.dumps(data),
                    content_type='application/json')
    res_data = json.loads(r.get_data(as_text=True))
    assert r.status_code == 400
    assert res_data['message']['since'] == "argument of type 'int' is not iterable"

    data = {'since': 'asdf'}
    r = client.post('/api/v1.0/waivers/+by-subjects-and-testcases', data=json.dumps(data),
                    content_type='application/json')
    res_data = json.loads(r.get_data(as_text=True))
    assert r.status_code == 400
    assert res_data['message']['since'] == \
        "time data 'asdf' does not match format '%Y-%m-%dT%H:%M:%S.%f'"


@pytest.mark.parametrize('trailing_slash', ('', '/'))
def test_about_endpoint(client, trailing_slash):
    r = client.get('/api/v1.0/about' + trailing_slash)
    assert r.status_code == 200

    output = json.loads(r.get_data(as_text=True))
    assert output['version'] == __version__
    assert output['auth_method'] == client.application.config['AUTH_METHOD']


def test_config_endpoint_permissions_map(client):
    config = {
        'PERMISSION_MAPPING': {
            '^kernel-qe\\.': {
                'groups': ['devel', 'qa'],
                'users': [],
            },
            '': {
                'groups': ['factory2-admins'],
                'users': [],
            },
        }
    }

    with patch.dict(client.application.config, config):
        r = client.get('/api/v1.0/config')

    assert r.status_code == 200
    assert r.json['permission_mapping'] == config['PERMISSION_MAPPING']


def test_permissions_endpoint(client):
    config = {
        'PERMISSIONS': [
            {
                "name": "^kernel-qe",
                "maintainers": ["alice@example.com"],
                "_testcase_regex_pattern": "^kernel-qe",
                "groups": ["devel", "qa"],
                "users": ["alice@example.com"],
            },
            {
                "name": "Greenwave Tests",
                "maintainers": ["greenwave-dev@example.com"],
                "testcases": ["greenwave-tests.*"],
                "groups": [],
                "users": ["HTTP/greenwave-dev.tests.example.com"]
            }
        ]
    }

    with patch.dict(client.application.config, config):
        r = client.get('/api/v1.0/permissions')
        assert r.status_code == 200
        assert r.json == config['PERMISSIONS']

        r = client.get('/api/v1.0/permissions?testcase=xxx')
        assert r.status_code == 200
        assert r.json == []

        r = client.get('/api/v1.0/permissions?testcase=greenwave-tests.test1')
        assert r.status_code == 200
        assert r.json == config['PERMISSIONS'][1:]

        r = client.get('/api/v1.0/permissions?testcase=kernel-qe.test1')
        assert r.status_code == 200
        assert r.json == config['PERMISSIONS'][0:1]


def test_config_endpoint_superusers(client):
    config = {
        'SUPERUSERS': ['alice', 'bob']
    }

    with patch.dict(client.application.config, config):
        r = client.get('/api/v1.0/config')

    assert r.status_code == 200
    assert r.json['superusers'] == config['SUPERUSERS']


def test_cors_good(client, session):
    headers = {
        'Access-Control-Request-Method': 'POST',
        'Access-Control-Request-Headers': 'Content-Type',
        'Origin': 'https://bodhi.fedoraproject.org',
    }
    r = client.options(
        '/api/v1.0/waivers/',
        content_type='Content-Type',
        headers=headers
    )

    assert r.status_code == 200
    assert r.headers.get('Access-Control-Allow-Origin') == 'https://bodhi.fedoraproject.org'
    assert 'POST' in r.headers.get('Access-Control-Allow-Methods', '').split(', ')


def test_cors_bad(client, session):
    headers = {
        'Access-Control-Request-Method': 'POST',
        'Access-Control-Request-Headers': 'Content-Type',
        'Origin': 'localhost',
    }
    r = client.options(
        '/api/v1.0/waivers/',
        content_type='Content-Type',
        headers=headers
    )

    assert r.status_code == 200
    assert 'Access-Control-Allow-Origin' not in r.headers
    assert 'Access-Control-Allow-Methods' not in r.headers


def test_create_multiple_waivers(mocked_user, client, session):
    item1 = {
        'subject_type': 'koji_build',
        'subject_identifier': 'glibc-2.26-27.fc27',
        'testcase': 'testcase1',
        'product_version': 'fool-1',
        'waived': True,
        'comment': 'it broke',
    }
    item2 = {
        'subject_type': 'koji_build',
        'subject_identifier': 'kernel-4.15.17-300.fc27',
        'testcase': 'testcase2',
        'product_version': 'fool-2',
        'waived': False,
        'comment': 'fixed',
    }
    data = [item1, item2]

    r = client.post('/api/v1.0/waivers/', data=json.dumps(data),
                    content_type='application/json')

    res_data = json.loads(r.get_data(as_text=True))
    assert r.status_code == 201
    assert isinstance(res_data, list)
    assert len(res_data) == 2

    actual_item1 = {k: v for k, v in res_data[0].items() if k in item1}
    actual_item2 = {k: v for k, v in res_data[1].items() if k in item2}
    assert actual_item1 == item1
    assert actual_item2 == item2

    # Transaction was not rolled back.
    assert session.query(Waiver).count() == 2


def test_create_multiple_waivers_rollback_on_error(mocked_user, client, session):
    item1 = {
        'subject_type': 'koji_build',
        'subject_identifier': 'glibc-2.26-27.fc27',
        'testcase': 'testcase1',
        'product_version': 'fool-1',
        'waived': True,
        'comment': 'it broke',
    }
    item2 = {}
    data = [item1, item2]

    r = client.post('/api/v1.0/waivers/', data=json.dumps(data),
                    content_type='application/json')

    assert r.status_code == 400

    # Transaction was rolled back.
    assert session.query(Waiver).count() == 0


def test_create_waiver_with_arbitrary_subject_type(mocked_user, client, session):
    data = {
        'subject_type': 'kind-of-magic',
        'subject_identifier': 'glibc-2.26-27.fc27',
        'testcase': 'testcase1',
        'product_version': 'fool-1',
        'waived': True,
        'comment': 'it broke',
    }
    r = client.post('/api/v1.0/waivers/', data=json.dumps(data),
                    content_type='application/json')
    res_data = json.loads(r.get_data(as_text=True))
    assert r.status_code == 201
    assert res_data['username'] == 'foo'
    assert res_data['subject'] == {'type': 'kind-of-magic', 'item': 'glibc-2.26-27.fc27'}
    assert res_data['subject_type'] == 'kind-of-magic'
    assert res_data['subject_identifier'] == 'glibc-2.26-27.fc27'
    assert res_data['testcase'] == 'testcase1'
    assert res_data['product_version'] == 'fool-1'
    assert res_data['waived'] is True
    assert res_data['comment'] == 'it broke'


def test_create_waiver_failed_event_once(mocked_user, client, session, caplog):
    config = dict(
        MESSAGE_BUS_PUBLISH=True,
        MESSAGE_PUBLISHER='stomp',
        MAX_STOMP_RETRY=3,
        STOMP_RETRY_DELAY_SECONDS=0,
        STOMP_CONFIGS={
            'destination': '/topic/VirtualTopic.eng.waiverdb.waiver.new',
            'connection': {
                'host_and_ports': [('broker01', 61612)],
            },
        },
    )

    data = {
        'subject_type': 'koji_build',
        'subject_identifier': 'glibc-2.26-27.fc27',
        'testcase': 'testcase1',
        'product_version': 'fool-1',
        'waived': True,
        'comment': 'it broke',
    }

    with patch.dict(client.application.config, config):
        with patch('waiverdb.events.stomp.Connection') as connection:
            connection().connect.side_effect = (StompException, StompException, None)
            r = client.post('/api/v1.0/waivers/', json=data)
            assert r.status_code == 201
            assert 'Failed to send message (try 1/3)' in caplog.text
            assert 'Failed to send message (try 2/3)' in caplog.text
            assert 'Failed to send message (try 3/3)' not in caplog.text
            assert 'StompException' in caplog.text



# SPDX-License-Identifier: GPL-2.0+

import os
from copy import copy
from mock import patch
import pytest
from sqlalchemy import create_engine
from waiverdb.app import create_app
from waiverdb.monitor import db_hook_event_listeners


@pytest.fixture(scope='session')
def app():
    os.environ['TEST'] = 'true'
    app = create_app()
    with app.app_context():
        yield app


@pytest.fixture(scope='session')
def db(app):
    """Session-wide test database."""
    from waiverdb.models import db
    dbname = db.engine.url.database
    # In order to drop and re-create the database, we have to connect to
    # template1 database in special AUTOCOMMIT isolation level.
    dburl = copy(db.engine.url)
    dburl.database = 'template1'
    if ':memory:' not in dbname:
        with create_engine(dburl).connect() as connection:
            connection.execution_options(isolation_level='AUTOCOMMIT')
            connection.execute('DROP DATABASE IF EXISTS {}'.format(dbname))
            connection.execute('CREATE DATABASE {}'.format(dbname))
    db.create_all()
    db_hook_event_listeners()
    return db


@pytest.fixture
def session(db, monkeypatch):
    """Patch Flask-SQLAlchemy to use a specific connection"""
    connection = db.engine.connect()
    transaction = connection.begin()

    # Patch Flask-SQLAlchemy to use our connection
    monkeypatch.setattr(db, 'get_engine', lambda *args: connection)

    yield db.session

    db.session.remove()
    transaction.rollback()
    connection.close()


@pytest.fixture
def client(app):
    """A Flask test client. An instance of :class:`flask.testing.TestClient`
    by default.
    """
    with app.test_client() as client:
        with patch('waiverdb.events.publish'):
            yield client


@pytest.fixture()
def enable_kerberos(app, monkeypatch):
    monkeypatch.setitem(app.config, 'AUTH_METHOD', 'Kerberos')


@pytest.fixture()
def enable_ssl(app, monkeypatch):
    monkeypatch.setitem(app.config, 'AUTH_METHOD', 'SSL')


@pytest.fixture()
def enable_ldap_host(app, monkeypatch):
    monkeypatch.setitem(app.config, 'LDAP_HOST', 'ldap://ldap.something.com')


@pytest.fixture()
def enable_ldap_base(app, monkeypatch):
    monkeypatch.setitem(app.config, 'LDAP_BASE', 'ou=Users,dc=something,dc=com')



# SPDX-License-Identifier: GPL-2.0+

import json
from base64 import b64encode

import mock
import ldap
import pytest
from click.testing import CliRunner
from textwrap import dedent
from werkzeug.exceptions import Unauthorized
from waiverdb.cli import cli as waiverdb_cli


@pytest.fixture()
def enable_permissions(app, monkeypatch):
    permissions = [
        {
            "testcases": ["testcase1.*"],
            "groups": ["factory-2-0"],
            "users": []
        },
        {
            "testcases": ["testcase2.*"],
            "groups": [],
            "users": ["foo"]
        },
        {
            "testcases": ["testcase3"],
            "groups": [],
            "users": []
        },
    ]
    monkeypatch.setitem(app.config, 'PERMISSIONS', permissions)


@pytest.mark.usefixtures('enable_permissions')
@pytest.mark.usefixtures('enable_kerberos')
@mock.patch.multiple("gssapi.SecurityContext", complete=True,
                     __init__=mock.Mock(return_value=None),
                     step=mock.Mock(return_value=b"STOKEN"),
                     initiator_name="foo@EXAMPLE.ORG")
@mock.patch.multiple("gssapi.Credentials",
                     __init__=mock.Mock(return_value=None),
                     __new__=mock.Mock(return_value=None))
class TestAccessControl(object):

    data = {
        'subject_type': 'koji_build',
        'subject_identifier': 'glibc-2.26-27.fc27',
        'testcase': 'testcase1.functional',
        'product_version': 'fool-1',
        'waived': True,
        'comment': 'it broke',
    }
    headers = {'Authorization':
               'Negotiate %s' % b64encode(b"CTOKEN").decode()}

    def test_ldap_host_base_not_defined(self, client, session):
        r = client.post('/api/v1.0/waivers/', data=json.dumps(self.data),
                        content_type='application/json', headers=self.headers)
        res_data = json.loads(r.get_data(as_text=True))
        assert r.status_code == 500
        assert res_data['message'] == ('LDAP_HOST and LDAP_SEARCHES also need to be defined '
                                       'if PERMISSIONS is defined.')

    @pytest.mark.usefixtures('enable_ldap_host')
    def test_ldap_host_defined_base_not(self, client, session):
        self.test_ldap_host_base_not_defined(client, session)

    @pytest.mark.usefixtures('enable_ldap_base')
    def test_ldap_base_defined_host_not(self, client, session):
        self.test_ldap_host_base_not_defined(client, session)

    @pytest.mark.usefixtures('enable_ldap_host')
    @pytest.mark.usefixtures('enable_ldap_base')
    @mock.patch('ldap.initialize', side_effect=ldap.LDAPError())
    def test_initialization_ldap_connection(self, mocked, client, session):
        r = client.post('/api/v1.0/waivers/', data=json.dumps(self.data),
                        content_type='application/json', headers=self.headers)
        res_data = json.loads(r.get_data(as_text=True))
        assert r.status_code == 401
        assert res_data['message'] == "Some error occurred initializing the LDAP connection."

    @pytest.mark.usefixtures('enable_ldap_host')
    @pytest.mark.usefixtures('enable_ldap_base')
    @mock.patch('waiverdb.authorization.get_group_membership', return_value=([]))
    def test_user_not_found_in_ldap(self, mocked_conn, client, session, monkeypatch):
        monkeypatch.setenv('KRB5_KTNAME', '/etc/foo.keytab')
        r = client.post('/api/v1.0/waivers/', data=json.dumps(self.data),
                        content_type='application/json', headers=self.headers)
        res_data = json.loads(r.get_data(as_text=True))
        assert r.status_code == 401
        assert res_data['message'] == "Couldn't find user foo in LDAP"

    @pytest.mark.usefixtures('enable_ldap_host')
    @pytest.mark.usefixtures('enable_ldap_base')
    @mock.patch('waiverdb.authorization.get_group_membership',
                return_value=(['factory-2-0', 'something-else']))
    def test_group_has_permission(self, mocked_conn, client, session, monkeypatch):
        monkeypatch.setenv('KRB5_KTNAME', '/etc/foo.keytab')
        r = client.post('/api/v1.0/waivers/', data=json.dumps(self.data),
                        content_type='application/json', headers=self.headers)
        res_data = json.loads(r.get_data(as_text=True))
        assert r.status_code == 201
        assert res_data['username'] == 'foo'
        assert res_data['subject'] == {'type': 'koji_build', 'item': 'glibc-2.26-27.fc27'}
        assert res_data['subject_type'] == 'koji_build'
        assert res_data['subject_identifier'] == 'glibc-2.26-27.fc27'
        assert res_data['testcase'] == 'testcase1.functional'
        assert res_data['product_version'] == 'fool-1'
        assert res_data['waived'] is True
        assert res_data['comment'] == 'it broke'

    @pytest.mark.usefixtures('enable_ldap_host')
    @pytest.mark.usefixtures('enable_ldap_base')
    def test_user_has_permission(self, client, session, monkeypatch):
        monkeypatch.setenv('KRB5_KTNAME', '/etc/foo.keytab')
        self.data['testcase'] = 'testcase2.integration'
        r = client.post('/api/v1.0/waivers/', data=json.dumps(self.data),
                        content_type='application/json', headers=self.headers)
        res_data = json.loads(r.get_data(as_text=True))
        assert r.status_code == 201
        assert res_data['username'] == 'foo'
        assert res_data['subject'] == {'type': 'koji_build', 'item': 'glibc-2.26-27.fc27'}
        assert res_data['subject_type'] == 'koji_build'
        assert res_data['subject_identifier'] == 'glibc-2.26-27.fc27'
        assert res_data['testcase'] == 'testcase2.integration'
        assert res_data['product_version'] == 'fool-1'
        assert res_data['waived'] is True
        assert res_data['comment'] == 'it broke'

    @pytest.mark.usefixtures('enable_ldap_host')
    @pytest.mark.usefixtures('enable_ldap_base')
    @mock.patch('waiverdb.authorization.get_group_membership',
                return_value=(['factory-2-0', 'something-else']))
    def test_both_user_group_no_permission(self, mocked_conn, client, session, monkeypatch):
        monkeypatch.setenv('KRB5_KTNAME', '/etc/foo.keytab')
        self.data['testcase'] = 'testcase3'
        r = client.post('/api/v1.0/waivers/', data=json.dumps(self.data),
                        content_type='application/json', headers=self.headers)
        res_data = json.loads(r.get_data(as_text=True))
        assert r.status_code == 401
        assert res_data['message'] == ("You are not authorized to submit a waiver "
                                       "for the test case testcase3")

    @pytest.mark.usefixtures('enable_ldap_host')
    @pytest.mark.usefixtures('enable_ldap_base')
    @mock.patch('waiverdb.auth.get_user', return_value=('bodhi', {}))
    @mock.patch('waiverdb.authorization.get_group_membership',
                return_value=(['factory-2-0', 'something-else']))
    def test_proxied_by_with_no_permission(self, mocked_conn, mock_get_user, client, session):
        self.data['testcase'] = 'testcase3'
        self.data['username'] = 'foo'
        r = client.post('/api/v1.0/waivers/', data=json.dumps(self.data),
                        content_type='application/json', headers=self.headers)
        res_data = json.loads(r.get_data(as_text=True))
        assert r.status_code == 401
        assert res_data['message'] == ("You are not authorized to submit a waiver "
                                       "for the test case testcase3")

    @pytest.mark.usefixtures('enable_ldap_host')
    @pytest.mark.usefixtures('enable_ldap_base')
    @mock.patch('waiverdb.auth.get_user', return_value=('bodhi', {}))
    @mock.patch('waiverdb.authorization.get_group_membership',
                return_value=(['factory-2-0', 'something-else']))
    def test_proxied_by_has_permission(self, mocked_conn, mock_get_user, client, session):
        self.data['testcase'] = 'testcase2.integration'
        self.data['username'] = 'foo'
        r = client.post('/api/v1.0/waivers/', data=json.dumps(self.data),
                        content_type='application/json', headers=self.headers)
        res_data = json.loads(r.get_data(as_text=True))
        assert r.status_code == 201
        assert res_data['username'] == 'foo'
        assert res_data['subject'] == {'type': 'koji_build', 'item': 'glibc-2.26-27.fc27'}
        assert res_data['subject_type'] == 'koji_build'
        assert res_data['subject_identifier'] == 'glibc-2.26-27.fc27'
        assert res_data['testcase'] == 'testcase2.integration'
        assert res_data['product_version'] == 'fool-1'
        assert res_data['waived'] is True
        assert res_data['comment'] == 'it broke'
        assert res_data['proxied_by'] == 'bodhi'

    @pytest.mark.usefixtures('enable_ldap_host')
    @pytest.mark.usefixtures('enable_ldap_base')
    @mock.patch('waiverdb.auth.get_user', return_value=('bodhi', {}))
    @mock.patch('requests.request',
                side_effect=Unauthorized('You are not authorized to submit a waiver for the test '
                                         'case testcase3'))
    def test_proxied_by_with_no_permission_cli(self, mock_request, mock_get_user, client, session,
                                               tmpdir):
        p = tmpdir.join('client.conf')
        p.write(dedent("""
            [waiverdb]
            auth_method=Kerberos
            api_url=http://localhost:5004/api/v1.0
            """))
        self.data['testcase'] = 'testcase3'
        self.data['username'] = 'foo'
        runner = CliRunner()
        args = ['-C', p.strpath, '-s', '{"type": "koji_build", "item": "glibc-2.26-27.fc27"}',
                '-t', 'testcase3', '-c', 'This is fine', '-p', 'fool-1', '-u', 'foo']
        result = runner.invoke(waiverdb_cli, args, catch_exceptions=True)
        assert result.exception.description == ("You are not authorized to submit a waiver "
                                                "for the test case testcase3")

    @pytest.mark.usefixtures('enable_ldap_host')
    @pytest.mark.usefixtures('enable_ldap_base')
    @mock.patch('waiverdb.auth.get_user', return_value=('bodhi', {}))
    def test_proxied_by_has_permission_cli(self, mock_get_user, client, session, tmpdir):
        with mock.patch('requests.request') as mock_request:
            mock_rv = mock.Mock()
            mock_rv.json.return_value = [{
                "comment": "it broke",
                "data": {"item": ["glibc-2.26-27.fc27"], "type": ["koji_build"]},
                "id": 15,
                "product_version": "fool-1",
                "testcase": "testcase2.integration",
                "timestamp": "2017-010-16T17:42:04.209638",
                "username": "foo",
                "proxied_by": "bodhi",
                "waived": True
            }]
            mock_request.return_value = mock_rv
            p = tmpdir.join('client.conf')
            p.write(dedent("""
                [waiverdb]
                auth_method=Kerberos
                api_url=http://localhost:5004/api/v1.0
                """))
            runner = CliRunner()
            args = ['-C', p.strpath, '-p', 'fool-1', '-r', '123',
                    '-c', "This is fine", '-u', 'foo']
            result = runner.invoke(waiverdb_cli, args)
            mock_request.assert_called_once()
            assert result.output.startswith('Created waiver 15 for result with id 123\n')



"""Add scenario field to the database

Revision ID: 3868a8118458
Revises: f6bc296ba966
Create Date: 2020-10-20 18:08:50.709804

"""

# revision identifiers, used by Alembic.
revision = '3868a8118458'
down_revision = 'f6bc296ba966'

from alembic import op
import sqlalchemy as sa


def upgrade():
    op.add_column('waiver', sa.Column('scenario', sa.String(length=255), nullable=True))


def downgrade():
    op.drop_column('waiver', 'scenario')



# SPDX-License-Identifier: GPL-2.0+
"""
CLI for creating new waivers against test results.
"""
import click
import requests
import json
import configparser
import re
from xmlrpc import client

from waiverdb import __version__


class OldJSONSubject(click.ParamType):
    """
    Deprecated JSON subject CLI parameter.
    """
    name = 'Deprecated JSON subject'

    def convert(self, value, param, ctx):
        if not isinstance(value, str):
            return value

        try:
            subject = json.loads(value)
        except json.JSONDecodeError as e:
            raise click.ClickException('Invalid JSON object: {}'.format(e))

        if not isinstance(subject, dict):
            raise click.ClickException(
                'Failed to parse JSON. Please use id and type instead of using subject.')

        return subject


def validate_config(config):
    """
    Validates the configuration needed for WaiverDB
    :return: None or ClickException
    """
    # Verify that all necessary config options are set
    config_error = ('The config option "{0}" is required')
    if not config.has_option('waiverdb', 'auth_method'):
        raise click.ClickException(config_error.format('auth_method'))
    auth_method = config.get('waiverdb', 'auth_method')
    if auth_method not in ['OIDC', 'Kerberos', 'dummy']:
        raise click.ClickException('The WaiverDB authentication mechanism of '
                                   '"{0}" is not supported'.format(auth_method))
    required_configs = ['api_url']
    if auth_method == 'OIDC':
        required_configs.append('oidc_id_provider')
        required_configs.append('oidc_client_id')
        required_configs.append('oidc_scopes')
    for required_config in required_configs:
        if not config.has_option('waiverdb', required_config):
            raise click.ClickException(config_error.format(required_config))


def print_result(waiver_id, result_description):
    click.echo('Created waiver {0} for result with {1}'.format(waiver_id, result_description))


def check_response(resp, result_ids):
    if not resp.ok:
        try:
            error_msg = resp.json()['message']
        except (ValueError, KeyError):
            error_msg = resp.text
        raise click.ClickException(
            'Failed to create waivers: {0}'
            .format(error_msg))

    response_data = resp.json()
    if result_ids:
        if len(response_data) != len(result_ids):
            raise RuntimeError(
                'Unexpected number of results in response: {!r}'.format(response_data))

        for result_id, data in zip(result_ids, response_data):
            waiver_id = data['id']
            msg = 'id {0}'.format(result_id)
            print_result(waiver_id, msg)
    else:
        for data in response_data:
            waiver_id = data['id']
            msg = 'subject type {0}, identifier {1} and testcase {2}'.format(
                data['subject_type'], data['subject_identifier'], data['testcase'])
            if data.get('scenario'):
                msg += f", scenario is {data['scenario']}"
            print_result(waiver_id, msg)


def guess_product_version(toparse, koji_build=False):
    if toparse == 'rawhide' or toparse.startswith('Fedora-Rawhide'):
        return 'fedora-rawhide'
    else:
        product_version = None
        if (toparse.startswith('f') and koji_build):
            product_version = 'fedora-'
        elif toparse.startswith('epel'):
            product_version = 'epel-'
        elif toparse.startswith('el'):
            product_version = 'rhel-'
        elif toparse.startswith('fc') or toparse.startswith('Fedora'):
            product_version = 'fedora-'
        if product_version:
            # seperate the prefix from the number
            result = list(filter(None, '-'.join(re.split(r'(\d+)', toparse)).split('-')))
            if len(result) >= 2:
                try:
                    int(result[1])
                    product_version += result[1]
                    return product_version
                except ValueError:
                    pass
    return None


@click.command(context_settings={'help_option_names': ['-h', '--help']})
@click.version_option(version=__version__)
@click.option('--config-file', '-C', default='/etc/waiverdb/client.conf',
              type=click.Path(exists=True),
              help='Specify a config file to use')
@click.option('--result-id', '-r', multiple=True, type=int,
              help='Specify one or more results to be waived')
@click.option('--scenario', '-S',
              help='Specify a scenario for a result to waive')
@click.option('--subject', '-s', type=OldJSONSubject(),
              help=('Deprecated. Use --subject-identifier and --subject-type instead. '
                    'Subject for a result to waive.'))
@click.option('--subject-identifier', '-i',
              help='Subject identifier for a result to waive.')
@click.option('--subject-type', '-T',
              help='Subject type for a result to waive.')
@click.option('--testcase', '-t',
              help='Specify a testcase for the subject')
@click.option('--product-version', '-p',
              help='Specify one of PDC\'s product version identifiers.')
@click.option('--waived/--no-waived', default=True,
              help='Whether or not the result is waived')
@click.option('--comment', '-c',
              help='A comment explaining why the result is waived')
@click.option('--username', '-u', default=None,
              help='Username on whose behalf the caller is proxying.')
def cli(username, comment, waived, product_version, testcase, scenario, subject,
        subject_identifier, subject_type, result_id, config_file):
    """
    Creates new waiver against test results.

    Examples:

    \b
        waiverdb-cli -r 47 -r 48 -p "fedora-28" -c "This is fine"

    \b
        waiverdb-cli -t dist.rpmdeplint -i qclib-1.3.1-3.fc28 -T bodhi_update \\
            -p "fedora-28" -c "This is expected for non-x86 packages"

    \b
        waiverdb-cli -t update.install_default_update_live -i FEDORA-2020-a70501de3d \\
            -T koji_build -S "fedora.updates-everything-boot-iso.x86_64.uefi" -c "This is ok"
    """
    config = configparser.ConfigParser()

    config.read(config_file)
    validate_config(config)

    result_ids = result_id

    # Backward compatibility with v0.11.0.
    if subject:
        if subject_identifier or subject_type:
            raise click.ClickException(
                'Please don\'t specify subject when using identifier and type.')
        if result_ids:
            raise click.ClickException('Please specify result_id or subject/testcase. Not both')

        subject_identifier = subject.get("productmd.compose.id")
        if subject_identifier:
            subject_type = "compose"
        else:
            subject_identifier = subject.get("item")
            subject_type = subject.get("type")

    if not comment:
        raise click.ClickException('Please specify comment')
    if result_ids and (testcase or subject_identifier):
        raise click.ClickException('Please specify result_id or id/type/testcase. Not both')
    if result_ids and scenario:
        raise click.ClickException('Please specify result_id or scenario. Not both')
    if not result_ids and not subject_identifier:
        raise click.ClickException('Please specify subject-identifier')
    if not result_ids and not testcase:
        raise click.ClickException('Please specify testcase')
    if not result_ids and not subject_type:
        raise click.ClickException('Please specify subject_type')

    if not product_version and not result_ids:
        # trying to guess the product_version
        if subject_type == 'koji_build':
            try:
                short_prod_version = subject_identifier.split('.')[-1]
                product_version = guess_product_version(short_prod_version, koji_build=True)
            except KeyError:
                pass

        # try to call koji
        if config.has_option('waiverdb', 'koji_base_url'):
            koji_base_url = config.get('waiverdb', 'koji_base_url')
            proxy = client.ServerProxy(koji_base_url)
            try:
                build = proxy.getBuild(subject_identifier)
                if build:
                    target = proxy.getTaskRequest(build['task_id'])[1]
                    product_version = guess_product_version(target, koji_build=True)
            except KeyError:
                pass
            except client.Fault:
                pass

        if subject_type == "compose":
            product_version = guess_product_version(subject_identifier)

    if not product_version:
        raise click.ClickException('Please specify product version using --product-version')

    auth_method = config.get('waiverdb', 'auth_method')
    data_list = []
    if not result_ids:
        data_list.append({
            'subject_identifier': subject_identifier,
            'subject_type': subject_type,
            'testcase': testcase,
            'waived': waived,
            'scenario': scenario,
            'product_version': product_version,
            'comment': comment,
            'username': username
        })

    # XXX - TODO - remove this in a future release.  (for backwards compat)
    for result_id in result_ids:
        data_list.append({
            'result_id': result_id,
            'waived': waived,
            'product_version': product_version,
            'comment': comment,
            'username': username
        })

    api_url = config.get('waiverdb', 'api_url')
    url = '{0}/waivers/'.format(api_url.rstrip('/'))
    data = json.dumps(data_list)
    common_request_arguments = {
        'data': data,
        'headers': {
            'Content-Type': 'application/json',
            'User-Agent': f'waiverdb-cli {__version__}',
        },
        'timeout': 60,
    }
    if auth_method == 'OIDC':
        # Try to import this now so the user gets immediate feedback if
        # it isn't installed
        try:
            import openidc_client  # noqa: F401
        except ImportError:
            raise click.ClickException('python-openidc-client needs to be installed')
        # Get the auth token using the OpenID client.
        oidc_client_secret = None
        if config.has_option('waiverdb', 'oidc_client_secret'):
            oidc_client_secret = config.get('waiverdb', 'oidc_client_secret')
        oidc = openidc_client.OpenIDCClient(
            'waiverdb',
            config.get('waiverdb', 'oidc_id_provider'),
            {'Token': 'Token', 'Authorization': 'Authorization'},
            config.get('waiverdb', 'oidc_client_id'),
            oidc_client_secret)
        scopes = config.get('waiverdb', 'oidc_scopes').strip().splitlines()

        resp = oidc.send_request(
            scopes=scopes,
            url=url,
            **common_request_arguments)
        check_response(resp, result_ids)
    elif auth_method == 'Kerberos':
        # Try to import this now so the user gets immediate feedback if
        # it isn't installed
        try:
            import requests_gssapi  # noqa: F401
        except ImportError:
            raise click.ClickException(
                'python-requests-gssapi needs to be installed')
        auth = requests_gssapi.HTTPKerberosAuth(
            mutual_authentication=requests_gssapi.OPTIONAL)
        resp = requests.request(
            'POST', url, auth=auth, **common_request_arguments)
        if resp.status_code == 401:
            failure_msg = ('WaiverDB authentication using GSSAPI failed. Make sure you have a '
                           'valid Kerberos ticket or that you correctly configured your Kerberos '
                           'configuration file. Please check the doc for troubleshooting '
                           'information.')
            failure_msg += '\nServer response: {}'.format(resp.text)
            try:
                msg = resp.json().get('message', failure_msg)
            except json.JSONDecodeError:
                msg = failure_msg
            raise click.ClickException(msg)
        check_response(resp, result_ids)
    elif auth_method == 'dummy':
        resp = requests.request(
            'POST', url, auth=('user', 'pass'), **common_request_arguments)
        check_response(resp, result_ids)


if __name__ == '__main__':
    cli()  # pylint: disable=E1120



"""waive absence of result

Revision ID: f2772c2c64a6
Revises: 0a74cdab732a
Create Date: 2017-12-04 10:03:54.792758

"""

from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = 'f2772c2c64a6'
down_revision = '0a74cdab732a'


def upgrade():
    op.add_column('waiver', sa.Column('subject', sa.Text(), nullable=True, index=True))
    op.add_column('waiver', sa.Column('testcase', sa.Text(), nullable=True, index=True))


def downgrade():
    op.drop_column('waiver', 'subject')
    op.drop_column('waiver', 'testcase')



# SPDX-License-Identifier: GPL-2.0+

"""This module contains tests for :mod:`waiverdb.events`."""
from __future__ import unicode_literals
from fedora_messaging import api, testing
from flask_restful import marshal
from waiverdb.models import Waiver
from waiverdb.fields import waiver_fields


def test_publish_new_waiver_with_fedmsg(session):
    waiver = Waiver(
        subject_type='koji_build',
        subject_identifier='glibc-2.26-27.fc27',
        testcase='testcase1',
        username='jcline',
        product_version='something',
        waived=True,
        comment='This is a comment',
    )

    sesh = session()
    sesh.add(waiver)
    sesh.flush()

    expected_msg = api.Message(
        topic='waiverdb.waiver.new',
        body=marshal(waiver, waiver_fields)
    )

    with testing.mock_sends(expected_msg):
        sesh.commit()


def test_publish_new_waiver_with_fedmsg_for_proxy_user(session):
    waiver = Waiver(
        subject_type='koji_build',
        subject_identifier='glibc-2.26-27.fc27',
        testcase='testcase1',
        username='jcline',
        product_version='something',
        waived=True,
        comment='This is a comment',
        proxied_by='bodhi'
    )
    sesh = session()
    sesh.add(waiver)
    sesh.flush()

    expected_msg = api.Message(
        topic='waiverdb.waiver.new',
        body=marshal(waiver, waiver_fields)
    )
    with testing.mock_sends(expected_msg):
        sesh.commit()



# SPDX-License-Identifier: GPL-2.0+

from base64 import b64encode
import pytest
import gssapi  # noqa
import mock
import json
from werkzeug.exceptions import Unauthorized
import waiverdb.auth
import flask_oidc


@pytest.mark.usefixtures('enable_kerberos')
class TestGSSAPIAuthentication(object):
    def test_unauthorized(self, client, monkeypatch):
        monkeypatch.setenv('KRB5_KTNAME', '/etc/foo.keytab')
        r = client.post('/api/v1.0/waivers/', content_type='application/json')
        assert r.status_code == 401
        assert r.headers.get('www-authenticate') == 'Negotiate'

    @mock.patch.multiple("gssapi.SecurityContext", complete=True,
                         __init__=mock.Mock(return_value=None),
                         step=mock.Mock(return_value=b"STOKEN"),
                         initiator_name="foo@EXAMPLE.ORG")
    @mock.patch.multiple("gssapi.Credentials",
                         __init__=mock.Mock(return_value=None),
                         __new__=mock.Mock(return_value=None))
    def test_authorized(self, client, monkeypatch):
        monkeypatch.setenv('KRB5_KTNAME', '/etc/foo.keytab')
        data = {
            'subject': {'type': 'koji_build', 'item': 'glibc-2.26-27.fc27'},
            'testcase': 'testcase1',
            'product_version': 'fool-1',
            'waived': True,
            'comment': 'it broke',
        }
        headers = {'Authorization':
                   'Negotiate %s' % b64encode(b"CTOKEN").decode()}
        r = client.post('/api/v1.0/waivers/', data=json.dumps(data),
                        content_type='application/json', headers=headers)
        assert r.status_code == 201
        assert r.headers.get('WWW-Authenticate') == \
            'negotiate %s' % b64encode(b"STOKEN").decode()
        res_data = json.loads(r.data.decode('utf-8'))
        assert res_data['username'] == 'foo'


class TestOIDCAuthentication(object):

    def test_get_user_without_token(self, session):
        with pytest.raises(Unauthorized) as excinfo:
            request = mock.MagicMock()
            waiverdb.auth.get_user(request)
        assert "No 'Authorization' header found" in excinfo.value.get_description()

    @mock.patch.object(flask_oidc.OpenIDConnect, '_get_token_info')
    def test_get_user_with_invalid_token(self, mocked_get_token, session):
        # http://vsbattles.wikia.com/wiki/Son_Goku
        name = 'Son Goku'
        mocked_get_token.return_value = {'active': False, 'username': name,
                                         'scope': 'openid waiverdb_scope'}
        headers = {'Authorization': 'Bearer invalid'}
        request = mock.MagicMock()
        request.headers.return_value = mock.MagicMock(spec_set=dict)
        request.headers.__getitem__.side_effect = headers.__getitem__
        request.headers.__setitem__.side_effect = headers.__setitem__
        request.headers.__contains__.side_effect = headers.__contains__
        with pytest.raises(Unauthorized) as excinfo:
            waiverdb.auth.get_user(request)
        assert 'Token required but invalid' in excinfo.value.get_description()

    @mock.patch.object(flask_oidc.OpenIDConnect, '_get_token_info')
    def test_get_user_good(self, mocked_get_token, session):
        # http://vsbattles.wikia.com/wiki/Son_Goku
        name = 'Son Goku'
        mocked_get_token.return_value = {'active': True, 'username': name,
                                         'scope': 'openid waiverdb_scope'}
        headers = {'Authorization': 'Bearer foobar'}
        request = mock.MagicMock()
        request.headers.return_value = mock.MagicMock(spec_set=dict)
        request.headers.__getitem__.side_effect = headers.__getitem__
        request.headers.__setitem__.side_effect = headers.__setitem__
        request.headers.__contains__.side_effect = headers.__contains__
        user, header = waiverdb.auth.get_user(request)
        assert user == name


@pytest.mark.usefixtures('enable_ssl')
class TestSSLAuthentication(object):
    def test_SSL_CLIENT_VERIFY_is_not_set_should_raise_error(self):
        with pytest.raises(Unauthorized) as excinfo:
            request = mock.MagicMock()
            waiverdb.auth.get_user(request)
        assert 'Cannot verify client' in excinfo.value.get_description()

    def test_SSL_CLIENT_S_DN_is_not_set_should_raise_error(self):
        with pytest.raises(Unauthorized) as excinfo:
            request = mock.MagicMock(environ={'SSL_CLIENT_VERIFY': 'SUCCESS'})
            waiverdb.auth.get_user(request)
        assert 'Unable to get user information (DN) from the client certificate' \
               in excinfo.value.get_description()

    def test_good_ssl_cert(self):
        # http://vsbattles.wikia.com/wiki/Son_Goku
        name = 'Son Goku'
        ssl = {
            'SSL_CLIENT_VERIFY': 'SUCCESS',
            'SSL_CLIENT_S_DN': name
        }
        request = mock.MagicMock(environ=ssl)
        user, header = waiverdb.auth.get_user(request)
        assert user == name



# SPDX-License-Identifier: GPL-2.0+

import logging
import sys


def log_to_stdout(level=logging.INFO):
    fmt = '%(asctime)s [pid %(process)5d] %(name)s %(levelname)s %(message)s'
    datefmt = '%Y-%m-%d %H:%M:%S'
    stream_handler = logging.StreamHandler(sys.stdout)
    stream_handler.setLevel(level)
    stream_handler.setFormatter(logging.Formatter(fmt=fmt, datefmt=datefmt))
    logging.getLogger().addHandler(stream_handler)


def init_logging(app):
    log_level = logging.DEBUG if app.debug else logging.INFO
    log_to_stdout(level=log_level)
    # In general we want to see everything from our own code,
    # but not detailed debug messages from third-party libraries.
    # Note that the log level on the handler above controls what
    # will actually appear on stdout.
    logging.getLogger().setLevel(logging.INFO)
    logging.getLogger('waiverdb').setLevel(logging.DEBUG)
    # The SQLALCHEMY_ECHO setting comes from Flask-SQLAlchemy, which translates
    # it to echo=True in the call to create_engine(). But SQLAlchemy itself
    # warns not to do that if you are configuring Python logging correctly:
    # http://docs.sqlalchemy.org/en/latest/core/engines.html#configuring-logging
    # We intercept that setting and do it "properly" instead.
    if app.config.get('SQLALCHEMY_ECHO'):
        logging.getLogger('sqlalchemy.engine').setLevel(logging.INFO)
        del app.config['SQLALCHEMY_ECHO']



# SPDX-License-Identifier: GPL-2.0+


import base64
import os
if not os.getenv('DOCS'):   # installing gssapi causing a problem for documentation building
    import gssapi
from flask import current_app, Response, g
from werkzeug.exceptions import Unauthorized, Forbidden


# Inspired by https://github.com/mkomitee/flask-kerberos/blob/master/flask_kerberos.py
# Later cleaned and ported to python-gssapi
def process_gssapi_request(token):
    try:
        stage = "initialize server context"
        sc = gssapi.SecurityContext(usage="accept")

        stage = "step context"
        token = sc.step(token if token != "" else None)
        token = token if token is not None else ""

        # The current architecture cannot support continuation here
        stage = "checking completion"
        if not sc.complete:
            current_app.logger.error(
                'Multiple GSSAPI round trips not supported')
            raise Forbidden("Attempted multiple GSSAPI round trips")

        current_app.logger.debug('Completed GSSAPI negotiation')

        stage = "getting remote user"
        user = str(sc.initiator_name)
        return user, token
    except gssapi.exceptions.GSSError as e:
        current_app.logger.exception(
            'Unable to authenticate: failed to %s: %s' %
            (stage, e.gen_message()))
        raise Forbidden("Authentication failed")


def get_user(request):
    user = None
    headers = dict()
    if current_app.config['AUTH_METHOD'] == 'OIDC':
        if 'Authorization' not in request.headers:
            raise Unauthorized("No 'Authorization' header found.")
        token = request.headers.get("Authorization").strip()
        prefix = 'Bearer '
        if not token.startswith(prefix):
            raise Unauthorized('Authorization headers must start with %r' % prefix)
        token = token[len(prefix):].strip()
        required_scopes = [
            'openid',
            current_app.config['OIDC_REQUIRED_SCOPE'],
        ]
        validity = current_app.oidc.validate_token(token, required_scopes)
        if validity is not True:
            raise Unauthorized(validity)
        user = g.oidc_token_info['username']
    elif current_app.config['AUTH_METHOD'] == 'Kerberos':
        if 'Authorization' not in request.headers:
            response = Response('Unauthorized', 401, {'WWW-Authenticate': 'Negotiate'})
            raise Unauthorized(response=response)
        header = request.headers.get("Authorization")
        token = ''.join(header.strip().split()[1:])
        user, token = process_gssapi_request(base64.b64decode(token))
        # remove realm
        user = user.split("@")[0]
        headers = {'WWW-Authenticate': ' '.join(
            ['negotiate', base64.b64encode(token).decode()])}
    elif current_app.config['AUTH_METHOD'] == 'SSL':
        # Nginx sets SSL_CLIENT_VERIFY and SSL_CLIENT_S_DN in request.environ
        # when doing SSL authentication.
        ssl_client_verify = request.environ.get('SSL_CLIENT_VERIFY')
        if ssl_client_verify != 'SUCCESS':
            raise Unauthorized('Cannot verify client: %s' % ssl_client_verify)
        if not request.environ.get('SSL_CLIENT_S_DN'):
            raise Unauthorized('Unable to get user information (DN) from the client certificate')
        user = request.environ.get('SSL_CLIENT_S_DN')
    elif current_app.config['AUTH_METHOD'] == 'dummy':
        # Blindly accept any username. For testing purposes only of course!
        if not request.authorization:
            response = Response('Unauthorized', 401, {'WWW-Authenticate': 'Basic realm="dummy"'})
            raise Unauthorized(response=response)
        user = request.authorization.username
    else:
        raise Unauthorized("Authenticated user required")
    return user, headers



# SPDX-License-Identifier: GPL-2.0+

# For an up-to-date version of this module, see:
#   https://pagure.io/monitor-flask-sqlalchemy

# pylint: disable=W,unexpected-keyword-arg,no-value-for-parameter

import os
import tempfile

from flask import Response
from flask.views import MethodView
from prometheus_client import (  # noqa: F401
    ProcessCollector, CollectorRegistry, Counter, multiprocess,
    Histogram, generate_latest, start_http_server, CONTENT_TYPE_LATEST)
from sqlalchemy import event

# Service-specific imports


if not os.environ.get('prometheus_multiproc_dir'):
    os.environ.setdefault('prometheus_multiproc_dir', tempfile.mkdtemp())
registry = CollectorRegistry()
ProcessCollector(registry=registry)
multiprocess.MultiProcessCollector(registry)
if os.getenv('MONITOR_STANDALONE_METRICS_SERVER_ENABLE', 'false') == 'true':
    port = os.getenv('MONITOR_STANDALONE_METRICS_SERVER_PORT', '10040')
    start_http_server(int(port), registry=registry)


# Generic metrics
messaging_tx_to_send_counter = Counter(
    'messaging_tx_to_send',
    'Total number of messages to send',
    registry=registry)
messaging_tx_stopped_counter = Counter(
    'messaging_tx_stopped',
    'Number of messages, which were eventually stopped before sending',
    registry=registry)
messaging_tx_sent_ok_counter = Counter(
    'messaging_tx_sent_ok',
    'Number of messages, which were sent successfully',
    registry=registry)
messaging_tx_failed_counter = Counter(
    'messaging_tx_failed',
    'Number of messages, for which the sender failed',
    registry=registry)

db_dbapi_error_counter = Counter(
    'db_dbapi_error',
    'Number of DBAPI errors',
    registry=registry)
db_engine_connect_counter = Counter(
    'db_engine_connect',
    'Number of \'engine_connect\' events',
    registry=registry)
db_handle_error_counter = Counter(
    'db_handle_error',
    'Number of exceptions during connection',
    registry=registry)
db_transaction_rollback_counter = Counter(
    'db_transaction_rollback',
    'Number of transactions, which were rolled back',
    registry=registry)

# Service-specific metrics
# XXX: TODO


def db_hook_event_listeners(target=None):
    # Service-specific import of db
    from waiverdb.models import db

    if not target:
        target = db.engine

    @event.listens_for(target, 'dbapi_error', named=True)
    def receive_dbapi_error(**kw):
        db_dbapi_error_counter.inc()

    @event.listens_for(target, 'engine_connect')
    def receive_engine_connect(conn, branch):
        db_engine_connect_counter.inc()

    @event.listens_for(target, 'handle_error')
    def receive_handle_error(exception_context):
        db_handle_error_counter.inc()

    @event.listens_for(target, 'rollback')
    def receive_rollback(conn):
        db_transaction_rollback_counter.inc()


class MonitorAPI(MethodView):
    def get(self):
        return Response(generate_latest(registry),
                        content_type=CONTENT_TYPE_LATEST)



# SPDX-License-Identifier: GPL-2.0+

import os


class Config(object):
    """
    A WaiverDB Flask configuration.
    """
    DEBUG = True
    DATABASE_URI = 'postgresql+psycopg2:///waiverdb'
    # We configure logging explicitly, turn off the Flask-supplied log handler.
    LOGGER_HANDLER_POLICY = 'never'
    HOST = '0.0.0.0'
    PORT = 5004
    PRODUCTION = False
    SHOW_DB_URI = False
    SECRET_KEY = 'replace-me-with-something-random'

    RESULTSDB_API_URL = 'https://taskotron.fedoraproject.org/resultsdb_api/api/v2.0'
    # need to explicitly turn this off
    # https://github.com/flask-restful/flask-restful/issues/449
    ERROR_404_HELP = False
    AUTH_METHOD = 'OIDC'  # Specify OIDC, Kerberos or SSL for authentication
    # Set this to True or False to enable publishing to a message bus
    MESSAGE_BUS_PUBLISH = True
    # Specify fedmsg or stomp for publishing messages
    MESSAGE_PUBLISHER = 'fedmsg'
    SQLALCHEMY_TRACK_MODIFICATIONS = True
    # A list of users are allowed to create waivers on behalf of other users.
    SUPERUSERS = []
    PERMISSIONS = []
    # Deprecated permission mapping
    PERMISSION_MAPPING = {}


class ProductionConfig(Config):
    DEBUG = False
    PRODUCTION = True


class DevelopmentConfig(Config):
    SQLALCHEMY_TRACK_MODIFICATIONS = True
    TRAP_BAD_REQUEST_ERRORS = True
    SHOW_DB_URI = True
    # The location of the client_secrets.json file used for API authentication
    OIDC_CLIENT_SECRETS = os.path.join(
        os.path.dirname(os.path.dirname(os.path.abspath(__file__))), 'conf',
        'client_secrets.json'
    )
    OIDC_REQUIRED_SCOPE = 'https://waiverdb.fedoraproject.org/oidc/create-waiver'
    OIDC_RESOURCE_SERVER_ONLY = True


class TestingConfig(Config):
    SQLALCHEMY_TRACK_MODIFICATIONS = True
    TRAP_BAD_REQUEST_ERRORS = True
    # Beware that the tests constantly wipe and re-create this database!
    # Do not configure this to point at any data you care about!
    DATABASE_URI = 'sqlite:///:memory:'
    TESTING = True
    OIDC_CLIENT_SECRETS = os.path.join(
        os.path.dirname(os.path.dirname(os.path.abspath(__file__))), 'tests',
        'client_secrets.json'
    )
    OIDC_REQUIRED_SCOPE = 'waiverdb_scope'
    OIDC_RESOURCE_SERVER_ONLY = True
    SUPERUSERS = ['bodhi']

    CORS_ORIGINS = 'https://bodhi.fedoraproject.org'



# SPDX-License-Identifier: GPL-2.0+

from flask_restful import fields

from waiverdb.models.waivers import subject_type_identifier_to_dict


class BackwardsCompatibleSubjectField(fields.Raw):

    def output(self, key, obj):
        waiver = obj
        return subject_type_identifier_to_dict(waiver.subject_type, waiver.subject_identifier)


waiver_fields = {
    'id': fields.Integer,
    'subject_type': fields.String,
    'subject_identifier': fields.String,
    'subject': BackwardsCompatibleSubjectField,
    'testcase': fields.String,
    'username': fields.String,
    'scenario': fields.String,
    'proxied_by': fields.String,
    'product_version': fields.String,
    'waived': fields.Boolean,
    'comment': fields.String,
    'timestamp': fields.DateTime(dt_format='iso8601'),
}



# SPDX-License-Identifier: GPL-2.0+
import pytest
import json
from mock import Mock, patch
from textwrap import dedent
from click.testing import CliRunner
from waiverdb import __version__
from waiverdb.cli import cli as waiverdb_cli
from waiverdb.cli import guess_product_version


def test_misconfigured_auth_method(tmpdir):
    p = tmpdir.join('client.conf')
    p.write("""
[waiverdb]
api_url=http://localhost:5004/api/v1.0
        """)
    runner = CliRunner()
    args = ['-C', p.strpath]
    result = runner.invoke(waiverdb_cli, args)
    assert result.exit_code == 1
    assert result.output == 'Error: The config option "auth_method" is required\n'


def test_misconfigured_api_url(tmpdir):
    p = tmpdir.join('client.conf')
    p.write("""
[waiverdb]
auth_method=OIDC
        """)
    runner = CliRunner()
    args = ['-C', p.strpath]
    result = runner.invoke(waiverdb_cli, args)
    assert result.exit_code == 1
    assert result.output == 'Error: The config option "api_url" is required\n'


def test_misconfigured_oidc_id_provider(tmpdir):
    p = tmpdir.join('client.conf')
    p.write("""
[waiverdb]
auth_method=OIDC
api_url=http://localhost:5004/api/v1.0
        """)
    runner = CliRunner()
    args = ['-C', p.strpath]
    result = runner.invoke(waiverdb_cli, args)
    assert result.exit_code == 1
    assert result.output == 'Error: The config option "oidc_id_provider" is required\n'


def test_misconfigured_oidc_client_id(tmpdir):
    p = tmpdir.join('client.conf')
    p.write("""
[waiverdb]
auth_method=OIDC
api_url=http://localhost:5004/api/v1.0
oidc_id_provider=https://id.stg.fedoraproject.org/openidc/
        """)
    runner = CliRunner()
    args = ['-C', p.strpath]
    result = runner.invoke(waiverdb_cli, args)
    assert result.exit_code == 1
    assert result.output == 'Error: The config option "oidc_client_id" is required\n'


def test_misconfigured_oidc_scopes(tmpdir):
    p = tmpdir.join('client.conf')
    p.write("""
[waiverdb]
auth_method=OIDC
api_url=http://localhost:5004/api/v1.0
oidc_id_provider=https://id.stg.fedoraproject.org/openidc/
oidc_client_id=waiverdb
        """)
    runner = CliRunner()
    args = ['-C', p.strpath]
    result = runner.invoke(waiverdb_cli, args)
    assert result.exit_code == 1
    assert result.output == 'Error: The config option "oidc_scopes" is required\n'


def test_no_product_version(tmpdir):
    p = tmpdir.join('client.conf')
    p.write("""
[waiverdb]
auth_method=OIDC
api_url=http://localhost:5004/api/v1.0
oidc_id_provider=https://id.stg.fedoraproject.org/openidc/
oidc_client_id=waiverdb
oidc_scopes=
    openid
        """)
    runner = CliRunner()
    args = ['-C', p.strpath, '-i', 'test', '-T', 'compose', '-t', 'testcase',
            '-c', 'comment']
    result = runner.invoke(waiverdb_cli, args)
    assert result.exit_code == 1
    assert result.output == 'Error: Please specify product version using --product-version\n'


def test_no_subject(tmpdir):
    p = tmpdir.join('client.conf')
    p.write("""
[waiverdb]
auth_method=OIDC
api_url=http://localhost:5004/api/v1.0
oidc_id_provider=https://id.stg.fedoraproject.org/openidc/
oidc_client_id=waiverdb
oidc_scopes=
    openid
        """)
    runner = CliRunner()
    args = ['-C', p.strpath, '-p', 'fedora-26', '-c', 'comment']
    result = runner.invoke(waiverdb_cli, args)
    assert result.exit_code == 1
    assert result.output == 'Error: Please specify subject-identifier\n'


def test_no_testcase(tmpdir):
    p = tmpdir.join('client.conf')
    p.write("""
[waiverdb]
auth_method=OIDC
api_url=http://localhost:5004/api/v1.0
oidc_id_provider=https://id.stg.fedoraproject.org/openidc/
oidc_client_id=waiverdb
oidc_scopes=
    openid
        """)
    runner = CliRunner()
    args = ['-C', p.strpath, '-p', 'fedora-26', '-i', 'item', '-T', 'compose', '-c', 'comment']
    result = runner.invoke(waiverdb_cli, args)
    assert result.exit_code == 1
    assert result.output == 'Error: Please specify testcase\n'


def test_no_comment(tmpdir):
    p = tmpdir.join('client.conf')
    p.write("""
[waiverdb]
auth_method=OIDC
api_url=http://localhost:5004/api/v1.0
oidc_id_provider=https://id.stg.fedoraproject.org/openidc/
oidc_client_id=waiverdb
oidc_scopes=
    openid
        """)
    runner = CliRunner()
    args = ['-C', p.strpath, '-p', 'fedora-26', '-i', 'item', '-T', 'compose', '-t', 'testcase']
    result = runner.invoke(waiverdb_cli, args)
    assert result.exit_code == 1
    assert result.output == 'Error: Please specify comment\n'


def test_oidc_auth_is_enabled(tmpdir):
    # Skip if waiverdb is rebuilt for an environment where GSSAPI
    # authentication is used and python-openidc-client is not available.
    pytest.importorskip('openidc_client')
    with patch('openidc_client.OpenIDCClient.send_request') as mock_oidc_req:
        mock_rv = Mock()
        mock_rv.json.return_value = [{
            "comment": "This is fine",
            "data": {"item": ["htop-1.0-1.fc22"], "type": ["bodhi_update"]},
            "id": 15,
            "product_version": "Parrot",
            "testcase": "test.testcase",
            "timestamp": "2017-010-16T17:42:04.209638",
            "username": "foo",
            "waived": True
        }]
        mock_oidc_req.return_value = mock_rv
        p = tmpdir.join('client.conf')
        p.write("""
[waiverdb]
auth_method=OIDC
api_url=http://localhost:5004/api/v1.0
oidc_id_provider=https://id.stg.fedoraproject.org/openidc/
oidc_client_id=waiverdb
oidc_scopes=
    openid
            """)
        runner = CliRunner()
        args = ['-C', p.strpath, '-p', 'Parrot', '-r', '123',
                '-c', "This is fine"]
        result = runner.invoke(waiverdb_cli, args)
        exp_json = [{
            "result_id": 123,
            "waived": True,
            "product_version": "Parrot",
            "comment": "This is fine",
            "username": None
        }]
        headers = {
            'Content-Type': 'application/json',
            'User-Agent': f'waiverdb-cli {__version__}',
        }
        mock_oidc_req.assert_called_once_with(
            url='http://localhost:5004/api/v1.0/waivers/',
            data=json.dumps(exp_json),
            scopes=['openid'],
            timeout=60,
            headers=headers)
        assert result.exit_code == 0
        assert result.output.startswith('Created waiver 15 for result with id 123\n')


def test_gssapi_is_enabled(tmpdir):
    with patch('requests.request') as mock_request:
        mock_rv = Mock()
        mock_rv.json.return_value = [{
            "comment": "This is fine",
            "data": {"item": ["htop-1.0-1.fc22"], "type": ["bodhi_update"]},
            "id": 15,
            "product_version": "Parrot",
            "testcase": "test.testcase",
            "timestamp": "2017-010-16T17:42:04.209638",
            "username": "foo",
            "waived": True
        }]
        mock_request.return_value = mock_rv
        p = tmpdir.join('client.conf')
        p.write("""
[waiverdb]
auth_method=Kerberos
api_url=http://localhost:5004/api/v1.0
            """)
        runner = CliRunner()
        args = ['-C', p.strpath, '-p', 'Parrot', '-r', '123',
                '-c', "This is fine"]
        result = runner.invoke(waiverdb_cli, args)
        mock_request.assert_called_once()
        assert result.output.startswith('Created waiver 15 for result with id 123\n')


def test_submit_waiver_with_id(tmpdir):
    with patch('requests.request') as mock_request:
        mock_rv = Mock()
        mock_rv.json.return_value = [{
            "comment": "This is fine",
            "data": {"item": ["htop-1.0-1.fc22"], "type": ["bodhi_update"]},
            "id": 15,
            "product_version": "Parrot",
            "testcase": {"name": "test.testcase"},
            "timestamp": "2017-010-16T17:42:04.209638",
            "username": "foo",
            "waived": True
        }]
        mock_request.return_value = mock_rv
        p = tmpdir.join('client.conf')
        p.write("""
[waiverdb]
auth_method=dummy
api_url=http://localhost:5004/api/v1.0
        """)
        runner = CliRunner()
        args = ['-C', p.strpath, '-p', 'Parrot', '-r', '123',
                '-c', "This is fine"]
        result = runner.invoke(waiverdb_cli, args, catch_exceptions=False)
        mock_request.assert_called()
        assert result.output == 'Created waiver 15 for result with id 123\n'


def test_submit_waiver_with_multiple_ids(tmpdir):
    with patch('requests.request') as mock_request:
        mock_rv = Mock()
        mock_rv.json.return_value = [
            {
                "comment": "This is fine",
                "data": {"item": ["htop-1.0-1.fc22"], "type": ["bodhi_update"]},
                "id": 15,
                "testcase": {"name": "test.testcase"},
                "timestamp": "2017-010-16T17:42:04.209638",
                "username": "foo",
                "waived": True
            }, {
                "comment": "This is fine",
                "data": {"item": ["htop-1.0-2.fc22"], "type": ["bodhi_update"]},
                "id": 16,
                "testcase": {"name": "test.testcase"},
                "timestamp": "2017-010-16T17:42:04.209638",
                "username": "foo",
                "waived": True
            }
        ]
        mock_request.return_value = mock_rv
        p = tmpdir.join('client.conf')
        p.write("""
[waiverdb]
auth_method=dummy
api_url=http://localhost:5004/api/v1.0
        """)
        runner = CliRunner()
        args = ['-C', p.strpath, '-p', 'Parrot', '-r', '123', '-r', '456',
                '-c', "This is fine"]
        result = runner.invoke(waiverdb_cli, args, catch_exceptions=False)
        mock_request.assert_called()

        assert result.output == 'Created waiver 15 for result with id 123\n\
Created waiver 16 for result with id 456\n'


def test_malformed_submission_with_id_and_subject_and_testcase(tmpdir):
    runner = CliRunner()
    p = tmpdir.join('client.conf')
    p.write("""
[waiverdb]
auth_method=dummy
api_url=http://localhost:5004/api/v1.0
    """)
    args = ['-C', p.strpath, '-p', 'Parrot', '-r', '123', '-s',
            '{"subject.test": "test", "s": "t"}', '-c', "This is fine"]
    result = runner.invoke(waiverdb_cli, args, catch_exceptions=False)
    assert result.output == 'Error: Please specify result_id or subject/testcase. Not both\n'


def test_malformed_submission_with_id_and_scenario(tmpdir):
    runner = CliRunner()
    p = tmpdir.join('client.conf')
    p.write(dedent("""
        [waiverdb]
        auth_method=dummy
        api_url=http://localhost:5004/api/v1.0
    """))
    args = ['-C', p.strpath, '-p', 'Parrot', '-r', '123', '-S', 'somescenario', '-c', "This is OK"]
    result = runner.invoke(waiverdb_cli, args, catch_exceptions=False)
    assert result.output == 'Error: Please specify result_id or scenario. Not both\n'


def test_submit_waiver_for_original_spec_nvr_result(tmpdir):
    with patch('requests.request') as mock_request:
        mock_rv = Mock()
        mock_rv.json.return_value = [{
            "comment": "This is fine",
            "original_spec_nvr": "test",
            "id": 15,
            "product_version": "Parrot",
            "testcase": {"name": "test.testcase"},
            "timestamp": "2017-010-16T17:42:04.209638",
            "username": "foo",
            "waived": True
        }]
        mock_request.return_value = mock_rv
        p = tmpdir.join('client.conf')
        p.write("""
[waiverdb]
auth_method=dummy
api_url=http://localhost:5004/api/v1.0
            """)
        runner = CliRunner()
        args = ['-C', p.strpath, '-p', 'Parrot', '-r', '123',
                '-c', "This is fine"]
        result = runner.invoke(waiverdb_cli, args)
        mock_request.assert_called()
        assert result.output == 'Created waiver 15 for result with id 123\n'


def test_create_waiver_product_version_missing(tmpdir):
    p = tmpdir.join('client.conf')
    p.write("""
[waiverdb]
auth_method=dummy
api_url=http://localhost:5004/api/v1.0
koji_base_url=https://koji.fedoraproject.org/kojihub
    """)
    runner = CliRunner()
    args = ['-C', p.strpath, '-s', '{"type": "koji_build", "item": "this-will-not-work"}',
            '-t', 'test.testcase', '-c', "This is fine"]
    result = runner.invoke(waiverdb_cli, args)
    assert result.output == 'Error: Please specify product version using --product-version\n'


def test_create_waiver_product_version_from_koji_build(tmpdir):
    with patch('requests.request') as mock_request:
        mock_rv = Mock()
        mock_rv.json.return_value = [{
            "comment": "This is fine",
            "id": 15,
            "subject_type": "koji_build",
            "subject_identifier": "setup-2.8.71-7.el7_4",
            "testcase": "test.testcase",
            "timestamp": "2017-010-16T17:42:04.209638",
            "username": "foo",
            "waived": True
        }]
        mock_request.return_value = mock_rv
        p = tmpdir.join('client.conf')
        p.write("""
[waiverdb]
auth_method=dummy
api_url=http://localhost:5004/api/v1.0
koji_base_url=https://koji.fedoraproject.org/kojihub
        """)
        runner = CliRunner()
        args = ['-C', p.strpath, '-s', '{"type": "koji_build", "item": "setup-2.8.71-7.el7_4"}',
                '-t', 'test.testcase', '-c', "This is fine"]
        result = runner.invoke(waiverdb_cli, args, catch_exceptions=False)
        mock_request.assert_called()
        assert result.output == (
            'Created waiver 15 for result with '
            'subject type koji_build, identifier setup-2.8.71-7.el7_4 '
            'and testcase test.testcase\n'
        )


def test_create_waiver_product_version_from_koji_build_scenario(tmpdir):
    with patch('requests.request') as mock_request:
        mock_rv = Mock()
        mock_rv.json.return_value = [{
            "comment": "This is fine",
            "id": 15,
            "subject_type": "koji_build",
            "subject_identifier": "setup-2.8.71-7.el7_4",
            "scenario": "somescenario",
            "testcase": "test.testcase",
            "timestamp": "2017-010-16T17:42:04.209638",
            "username": "foo",
            "waived": True
        }]
        mock_request.return_value = mock_rv
        p = tmpdir.join('client.conf')
        p.write(dedent("""
            [waiverdb]
            auth_method=dummy
            api_url=http://localhost:5004/api/v1.0
            koji_base_url=https://koji.fedoraproject.org/kojihub
        """))
        runner = CliRunner()
        args = [
            '-C', p.strpath, '-s', '{"type": "koji_build", "item": "setup-2.8.71-7.el7_4"}',
            '-S', 'somescenario', '-t', 'test.testcase', '-c', "This is fine"
        ]
        result = runner.invoke(waiverdb_cli, args, catch_exceptions=False)
        mock_request.assert_called()
        assert result.output == (
            'Created waiver 15 for result with '
            'subject type koji_build, identifier setup-2.8.71-7.el7_4 '
            'and testcase test.testcase, scenario is somescenario\n'
        )


def test_create_waiver_product_version_from_compose(tmpdir):
    with patch('requests.request') as mock_request:
        mock_rv = Mock()
        mock_rv.json.return_value = [{
            "comment": "This is fine",
            "id": 15,
            "subject_type": "compose",
            "subject_identifier": "Fedora-Rawhide-20180526.n.1",
            "testcase": "test.testcase",
            "timestamp": "2017-010-16T17:42:04.209638",
            "username": "foo",
            "waived": True
        }]
        mock_request.return_value = mock_rv
        p = tmpdir.join('client.conf')
        p.write("""
[waiverdb]
auth_method=dummy
api_url=http://localhost:5004/api/v1.0
koji_base_url=https://koji.fedoraproject.org/kojihub
        """)
        runner = CliRunner()
        args = ['-C', p.strpath, '-s', ('{"productmd.compose.id": "Fedora-Rawhide-20180526.n.1",'
                                        '"type": "compose",'
                                        '"item": "Fedora-Rawhide-20180526.n.1"}'),
                '-t', 'test.testcase', '-c', "This is fine"]
        result = runner.invoke(waiverdb_cli, args)
        mock_request.assert_called()
        assert result.output == (
            'Created waiver 15 for result with '
            'subject type compose, identifier Fedora-Rawhide-20180526.n.1 '
            'and testcase test.testcase\n'
        )


def test_guess_product_version():
    # some more tests for checking "guess_product_version"
    assert guess_product_version('epel7-infra-mailman') == 'epel-7'
    assert guess_product_version('f26-infra', koji_build=True) == 'fedora-26'
    assert guess_product_version('Fedora-28-20180423.n.0') == 'fedora-28'
    assert guess_product_version('Fedora-Rawhide-20180524.n.0') == 'fedora-rawhide'
    assert guess_product_version('Fedora-Atomic-28-20180424.4') is None


def test_create_waiver_missing_subject_identifier(tmpdir):
    p = tmpdir.join('client.conf')
    p.write("""
[waiverdb]
auth_method=dummy
api_url=http://localhost:5004/api/v1.0
koji_base_url=https://koji.fedoraproject.org/kojihub
    """)
    runner = CliRunner()
    args = ['-C', p.strpath, '-T', 'koji_build',
            '-t', 'test.testcase', '-c', "This is fine"]
    result = runner.invoke(waiverdb_cli, args)
    assert result.exit_code != 0
    assert 'Error: Please specify subject-identifier' in result.output


def test_create_waiver_missing_subject_type(tmpdir):
    p = tmpdir.join('client.conf')
    p.write("""
[waiverdb]
auth_method=dummy
api_url=http://localhost:5004/api/v1.0
koji_base_url=https://koji.fedoraproject.org/kojihub
    """)
    runner = CliRunner()
    args = ['-C', p.strpath, '-i', 'setup-2.8.71-7.el7_4',
            '-t', 'test.testcase', '-c', "This is fine"]
    result = runner.invoke(waiverdb_cli, args)
    assert result.exit_code != 0
    assert 'Error: Please specify subject_type' in result.output


def test_submit_waiver_with_arbitrary_subject_type(tmpdir):
    with patch('requests.request') as mock_request:
        mock_rv = Mock()
        mock_rv.json.return_value = [{
            "comment": "This is fine",
            "id": 15,
            "product_version": "Parrot",
            "subject_type": "some-kind-of-magic",
            "subject_identifier": "setup-2.8.71-7.el7_4",
            "testcase": "test.testcase",
            "timestamp": "2017-010-16T17:42:04.209638",
            "username": "foo",
            "waived": True
        }]
        mock_request.return_value = mock_rv
        p = tmpdir.join('client.conf')
        p.write("""
[waiverdb]
auth_method=dummy
api_url=http://localhost:5004/api/v1.0
            """)
        runner = CliRunner()
        args = ['-C', p.strpath, '-p', 'Parrot',
                '-s', '{"type": "some-kind-of-magic", "item": "setup-2.8.71-7.el7_4"}',
                '-t', 'test.testcase', '-c', "This is fine"]
        result = runner.invoke(waiverdb_cli, args)
        mock_request.assert_called()
        assert result.output == ('Created waiver 15 for result with subject type '
                                 'some-kind-of-magic, identifier setup-2.8.71-7.el7_4 and '
                                 'testcase test.testcase\n')


def test_waiver_cli_version():
    runner = CliRunner()
    args = ['--version']
    result = runner.invoke(waiverdb_cli, args)
    assert result.output == f'cli, version {__version__}\n'
    assert result.exit_code == 0



# SPDX-License-Identifier: GPL-2.0+

import datetime
from .base import db
from sqlalchemy import or_, and_, false


def subject_dict_to_type_identifier(subject):
    """
    WaiverDB < 0.11 accepted an arbitrary dict for the 'subject'.
    Now we expect a specific type and identifier.
    This maps from the old style to the new, for backwards compatibility.
    """
    # handling the special cases...
    if (subject.get('type') in ['koji_build', 'brew-build']
            and 'item' in subject
            and isinstance(subject['item'], str)):
        return ('koji_build', subject['item'])
    elif 'original_spec_nvr' in subject and isinstance(subject['original_spec_nvr'], str):
        return ('koji_build', subject['original_spec_nvr'])
    elif 'productmd.compose.id' in subject and isinstance(subject['productmd.compose.id'], str):
        return ('compose', subject['productmd.compose.id'])
    # then handling the general case...
    elif 'item' in subject and isinstance(subject['item'], str):
        return (subject.get('type'), subject['item'])
    else:
        raise ValueError('Subject type should be non empty string, actual value is: %r' % subject)


def subject_type_identifier_to_dict(subject_type, subject_identifier):
    """
    Inverse of the above function.
    This is for backwards compatibility in *responses*.
    """
    if subject_type == 'compose':
        return {'productmd.compose.id': subject_identifier}
    elif subject_type and isinstance(subject_type, str):
        return {'type': subject_type, 'item': subject_identifier}
    else:
        raise ValueError(('Subject type should be non empty string, '
                          'actual value is: %r') % subject_type)


class Waiver(db.Model):
    id = db.Column(db.Integer, primary_key=True)
    subject_type = db.Column(db.Text, nullable=False, index=True)
    subject_identifier = db.Column(db.Text, nullable=False, index=True)
    testcase = db.Column(db.Text, nullable=False, index=True)
    username = db.Column(db.String(255), nullable=False)
    proxied_by = db.Column(db.String(255))
    product_version = db.Column(db.String(200), nullable=False)
    waived = db.Column(db.Boolean, nullable=False, default=False)
    scenario = db.Column(db.String(255), nullable=True)
    comment = db.Column(db.Text)
    timestamp = db.Column(db.DateTime, default=datetime.datetime.utcnow)
    __table_args__ = (
        db.Index('ix_waiver_subject_type_identifier', subject_type, subject_identifier),
    )

    def __init__(self, subject_type, subject_identifier, testcase, username, product_version,
                 waived=False, comment=None, proxied_by=None, scenario=None):
        self.subject_type = subject_type
        self.subject_identifier = subject_identifier
        self.testcase = testcase
        self.username = username
        self.product_version = product_version
        self.waived = waived
        self.comment = comment
        self.proxied_by = proxied_by
        self.scenario = scenario

    def __repr__(self):
        return ('%s(subject_type=%r, subject_identifier=%r, testcase=%r, scenario=%r, username=%r, '
                'product_version=%r, waived=%r)'
                % (self.__class__.__name__, self.subject_type, self.subject_identifier,
                   self.testcase, self.scenario, self.username, self.product_version, self.waived))

    @classmethod
    def by_results(cls, query, results):
        """
        Filter ``query`` by matching with at least one filter in ``results``.

        If ``results`` is empty, ``query`` is not filtered.

        Args:
            query (flask_sqlalchemy.BaseQuery)
            results (list): each item should be dict containing
                "subject" (dict) and "testcase" (str), both optional

        Returns:
            Filtered query.
        """
        clauses = []
        for result in results:
            subject = result.get('subject', None)
            testcase = result.get('testcase', None)
            if not subject and not testcase:
                continue
            inner_clauses = []
            if subject:
                try:
                    subject_type, subject_identifier = subject_dict_to_type_identifier(subject)
                except ValueError:
                    inner_clauses.append(false())
                else:
                    inner_clauses.append(cls.subject_type == subject_type)
                    inner_clauses.append(cls.subject_identifier == subject_identifier)
            if testcase:
                inner_clauses.append(cls.testcase == testcase)
            clauses.append(and_(*inner_clauses))

        return query.filter(or_(*clauses))



# SPDX-License-Identifier: GPL-2.0+

"""This module contains tests for :mod:`waiverdb.app`."""
from __future__ import unicode_literals

import mock

from waiverdb import app, config
from flask_sqlalchemy import SignallingSession


class DisabledMessagingConfig(config.Config):
    MESSAGE_BUS_PUBLISH = False
    AUTH_METHOD = None
    SQLALCHEMY_TRACK_MODIFICATIONS = True


class EnabledMessagedConfig(config.Config):
    MESSAGE_BUS_PUBLISH = True
    AUTH_METHOD = None
    SQLALCHEMY_TRACK_MODIFICATIONS = True


@mock.patch('waiverdb.app.event.listen')
def test_disabled_messaging_should_not_register_events(mock_listen):
    app.create_app(DisabledMessagingConfig)
    assert 0 == mock_listen.call_count


@mock.patch('waiverdb.app.event.listen')
def test_enabled_messaging_should_register_events(mock_listen):
    app.create_app(EnabledMessagedConfig)
    mock_listen.assert_called_once_with(
        SignallingSession, 'after_commit', app.publish_new_waiver)



"""set nullable on new and old fields.

Revision ID: ed43eb9b221c
Revises: 71b84ccc31bb
Create Date: 2018-02-14 12:09:42.877375

"""

# revision identifiers, used by Alembic.
revision = 'ed43eb9b221c'
down_revision = '71b84ccc31bb'

from alembic import op


def upgrade():
    op.alter_column('waiver', 'subject', nullable=False)
    op.alter_column('waiver', 'testcase', nullable=False)
    op.alter_column('waiver', 'result_id', nullable=True)


def downgrade():
    op.alter_column('waiver', 'subject', nullable=True)
    op.alter_column('waiver', 'testcase', nullable=True)
    op.alter_column('waiver', 'result_id', nullable=False)



from waiverdb.api_v1 import permissions
from waiverdb.authorization import match_testcase_permissions


def test_permissions_mapping_compat(app, monkeypatch):
    """
    Verify backwards compatibility with deprecated PERMISSION_MAPPING option.
    """
    monkeypatch.setitem(app.config, 'PERMISSIONS', [])
    monkeypatch.setitem(app.config, 'PERMISSION_MAPPING', {})
    assert permissions() == []

    permission_mapping = {
        "^kernel-qe": {
            "groups": ["devel", "qa"],
            "users": ["alice@example.com"],
            "maintainer": "alice@example.com",
        },
    }
    monkeypatch.setitem(app.config, 'PERMISSION_MAPPING', permission_mapping)
    assert permissions() == [
        {
            "name": "^kernel-qe",
            "maintainers": ["alice@example.com"],
            "_testcase_regex_pattern": "^kernel-qe",
            "groups": ["devel", "qa"],
            "users": ["alice@example.com"],
        }
    ]

    permissions_override = [
        {
            "name": "Greenwave Tests",
            "maintainers": ["greenwave-dev@example.com"],
            "testcases": ["greenwave-tests.*"],
            "groups": [],
            "users": ["HTTP/greenwave-dev.tests.example.com"]
        }
    ]
    monkeypatch.setitem(app.config, 'PERMISSIONS', permissions_override)
    assert permissions() == permissions_override


def test_match_testcase_permissions():
    """
    Verify that correct permissions are retrieved for given test case name.
    """
    assert list(match_testcase_permissions("kernel-qe.test1", [])) == []

    permissions = [
        {
            "name": "^kernel-qe",
            "maintainers": ["alice@example.com"],
            "_testcase_regex_pattern": "^kernel-qe",
            "groups": ["devel", "qa"],
            "users": ["alice@example.com"],
        },
        {
            "name": "Greenwave Tests",
            "maintainers": ["greenwave-dev@example.com"],
            "testcases": ["greenwave-tests.*"],
            "groups": [],
            "users": ["HTTP/greenwave-dev.tests.example.com"]
        }
    ]
    assert list(match_testcase_permissions("greenwave-tests.test1", permissions)) \
        == [permissions[1]]
    assert list(match_testcase_permissions("kernel-qe.test1", permissions)) \
        == [permissions[0]]



# SPDX-License-Identifier: GPL-2.0+

import os
import pytest
import requests
import waiverdb.monitor

from six.moves import reload_module


def test_metrics(client):
    r = client.get('/api/v1.0/metrics')

    assert r.status_code == 200
    assert len([line for line in r.get_data(as_text=True).splitlines()
                if line.startswith('# TYPE messaging_')
                and line.endswith(' counter')]) == 4
    assert len([line for line in r.get_data(as_text=True).splitlines()
                if line.startswith('# TYPE db_')
                and line.endswith(' counter')]) == 4


def test_standalone_metrics_server_disabled_by_default():
    with pytest.raises(requests.exceptions.ConnectionError):
        requests.get('http://127.0.0.1:10040/metrics')


def test_standalone_metrics_server():
    os.environ['MONITOR_STANDALONE_METRICS_SERVER_ENABLE'] = 'true'
    reload_module(waiverdb.monitor)

    r = requests.get('http://127.0.0.1:10040/metrics')

    assert len([line for line in r.text.splitlines()
                if line.startswith('# TYPE messaging_')
                and line.endswith(' counter')]) == 4
    assert len([line for line in r.text.splitlines()
                if line.startswith('# TYPE db_')
                and line.endswith(' counter')]) == 4



# SPDX-License-Identifier: GPL-2.0+

"""Replace waiver.subject dict with subject_type and subject_identifier

Revision ID: f6bc296ba966
Revises: ce8a1351ecdc
Create Date: 2018-04-24 16:21:42.017640
"""

# revision identifiers, used by Alembic.
revision = 'f6bc296ba966'
down_revision = 'ce8a1351ecdc'

from alembic import op
from sqlalchemy import Column, Text, Integer, null
# These are the "lightweight" SQL expression versions (not using metadata):
from sqlalchemy.sql.expression import table, column, select, update
from waiverdb.models.base import EqualityComparableJSONType
from waiverdb.models.waivers import subject_dict_to_type_identifier, \
    subject_type_identifier_to_dict

def upgrade():
    # Create the columns NULLable first, so that we can populate them
    op.add_column('waiver', Column('subject_identifier', Text, nullable=True))
    op.add_column('waiver', Column('subject_type', Text, nullable=True))

    # Lightweight table definition for producing queries:
    waiver_table = table('waiver',
                         column('id', type_=Integer),
                         column('subject', type_=EqualityComparableJSONType),
                         column('subject_type', type_=Text),
                         column('subject_identifier', type_=Text))

    # Fill in values for the new columns
    connection = op.get_bind()
    rows = connection.execute(select([waiver_table.c.id, waiver_table.c.subject]))
    for waiver_id, subject in rows:
        try:
            subject_type, subject_identifier = subject_dict_to_type_identifier(subject)
        except ValueError:
            # The 'subject' value might be invalid, see: https://pagure.io/waiverdb/issue/210
            # Let's map it to something which is valid but will never match
            # anything Greenwave is looking for.
            # Note that the original, invalid 'subject' value is still
            # preserved in the row in case of downgrade. So we are not losing
            # any data here.
            subject_type, subject_identifier = 'koji_build', ''
        connection.execute(update(waiver_table)
                           .where(waiver_table.c.id == waiver_id)
                           .values(subject_type=subject_type,
                                   subject_identifier=subject_identifier))

    # Now make the columns non-NULLable, and populate indexes
    op.alter_column('waiver', 'subject_type', nullable=False)
    op.alter_column('waiver', 'subject_identifier', nullable=False)
    op.create_index('ix_waiver_subject_identifier', 'waiver', ['subject_identifier'])
    op.create_index('ix_waiver_subject_type', 'waiver', ['subject_type'])
    op.create_index('ix_waiver_subject_type_identifier', 'waiver', ['subject_type', 'subject_identifier'])

    # Make the old column NULLable for newly inserted rows
    # (we keep it around in case of downgrade though)
    op.alter_column('waiver', 'subject', nullable=True)


def downgrade():
    # Lightweight table definition for producing queries:
    waiver_table = table('waiver',
                         column('id', type_=Integer),
                         column('subject', type_=EqualityComparableJSONType),
                         column('subject_type', type_=Text),
                         column('subject_identifier', type_=Text))

    # Fill in the old column for any waivers inserted since the upgrade
    connection = op.get_bind()
    rows = connection.execute(
            select([waiver_table.c.id, waiver_table.c.subject_type, waiver_table.c.subject_identifier])
            .where(waiver_table.c.subject.is_(null())))
    for waiver_id, subject_type, subject_identifier in rows:
        subject = subject_type_identifier_to_dict(subject_type, subject_identifier)
        connection.execute(update(waiver_table)
                           .where(waiver_table.c.id == waiver_id)
                           .values(subject=subject))

    # Drop the new columns
    op.drop_index('ix_waiver_subject_type_identifier', table_name='waiver')
    op.drop_index('ix_waiver_subject_type', table_name='waiver')
    op.drop_index('ix_waiver_subject_identifier', table_name='waiver')
    op.drop_column('waiver', 'subject_type')
    op.drop_column('waiver', 'subject_identifier')

    # Make the old column non-NULLable again
    op.alter_column('waiver', 'subject', nullable=False)
